"""Code generated by Speakeasy (https://speakeasy.com). DO NOT EDIT."""

from __future__ import annotations
from .dto_overridelineitemrequest import (
    DtoOverrideLineItemRequest,
    DtoOverrideLineItemRequestTypedDict,
)
from flexprice.types import BaseModel, UNSET_SENTINEL
from pydantic import model_serializer
from typing import Dict, List, Optional
from typing_extensions import NotRequired, TypedDict


class DtoSubscriptionPhaseCreateRequestTypedDict(TypedDict):
    start_date: str
    coupons: NotRequired[List[str]]
    r"""Coupons represents subscription-level coupons to be applied to this phase"""
    end_date: NotRequired[str]
    line_item_coupons: NotRequired[Dict[str, List[str]]]
    r"""LineItemCoupons represents line item-level coupons (map of line_item_id to coupon IDs)"""
    metadata: NotRequired[Dict[str, str]]
    override_line_items: NotRequired[List[DtoOverrideLineItemRequestTypedDict]]
    r"""OverrideLineItems allows customizing specific prices for this phase
    If not provided, phase will use the same line items as the subscription (plan prices)
    """


class DtoSubscriptionPhaseCreateRequest(BaseModel):
    start_date: str

    coupons: Optional[List[str]] = None
    r"""Coupons represents subscription-level coupons to be applied to this phase"""

    end_date: Optional[str] = None

    line_item_coupons: Optional[Dict[str, List[str]]] = None
    r"""LineItemCoupons represents line item-level coupons (map of line_item_id to coupon IDs)"""

    metadata: Optional[Dict[str, str]] = None

    override_line_items: Optional[List[DtoOverrideLineItemRequest]] = None
    r"""OverrideLineItems allows customizing specific prices for this phase
    If not provided, phase will use the same line items as the subscription (plan prices)
    """

    @model_serializer(mode="wrap")
    def serialize_model(self, handler):
        optional_fields = set(
            [
                "coupons",
                "end_date",
                "line_item_coupons",
                "metadata",
                "override_line_items",
            ]
        )
        serialized = handler(self)
        m = {}

        for n, f in type(self).model_fields.items():
            k = f.alias or n
            val = serialized.get(k, serialized.get(n))

            if val != UNSET_SENTINEL:
                if val is not None or k not in optional_fields:
                    m[k] = val

        return m



"""Code generated by Speakeasy (https://speakeasy.com). DO NOT EDIT."""

from __future__ import annotations
from flexprice.types import BaseModel, UNSET_SENTINEL
from flexprice.utils import FieldMetadata, PathParamMetadata, QueryParamMetadata
from pydantic import model_serializer
from typing import Optional
from typing_extensions import Annotated, NotRequired, TypedDict


class GetInvoicePdfRequestTypedDict(TypedDict):
    id: str
    r"""Invoice ID"""
    url: NotRequired[bool]
    r"""Return presigned URL from s3 instead of PDF"""


class GetInvoicePdfRequest(BaseModel):
    id: Annotated[
        str, FieldMetadata(path=PathParamMetadata(style="simple", explode=False))
    ]
    r"""Invoice ID"""

    url: Annotated[
        Optional[bool],
        FieldMetadata(query=QueryParamMetadata(style="form", explode=True)),
    ] = None
    r"""Return presigned URL from s3 instead of PDF"""

    @model_serializer(mode="wrap")
    def serialize_model(self, handler):
        optional_fields = set(["url"])
        serialized = handler(self)
        m = {}

        for n, f in type(self).model_fields.items():
            k = f.alias or n
            val = serialized.get(k, serialized.get(n))

            if val != UNSET_SENTINEL:
                if val is not None or k not in optional_fields:
                    m[k] = val

        return m



"""Code generated by Speakeasy (https://speakeasy.com). DO NOT EDIT."""

from __future__ import annotations
from flexprice.types import BaseModel
from flexprice.utils import FieldMetadata, QueryParamMetadata
from typing_extensions import Annotated, TypedDict


class GetTaskResultRequestTypedDict(TypedDict):
    workflow_id: str
    r"""Workflow ID"""


class GetTaskResultRequest(BaseModel):
    workflow_id: Annotated[
        str, FieldMetadata(query=QueryParamMetadata(style="form", explode=True))
    ]
    r"""Workflow ID"""



"""Code generated by Speakeasy (https://speakeasy.com). DO NOT EDIT."""

from __future__ import annotations
from .dto_integrationentitymapping import (
    DtoIntegrationEntityMapping,
    DtoIntegrationEntityMappingTypedDict,
)
from .dto_taxrateoverride import DtoTaxRateOverride, DtoTaxRateOverrideTypedDict
from flexprice.types import BaseModel, UNSET_SENTINEL
from pydantic import model_serializer
from typing import Dict, List, Optional
from typing_extensions import NotRequired, TypedDict


class DtoCreateCustomerRequestTypedDict(TypedDict):
    r"""Request object for creating a new customer in the system"""

    external_id: str
    r"""external_id is the unique identifier from your system to reference this customer (required)"""
    address_city: NotRequired[str]
    r"""address_city is the city name with maximum 100 characters"""
    address_country: NotRequired[str]
    r"""address_country is the two-letter ISO 3166-1 alpha-2 country code"""
    address_line1: NotRequired[str]
    r"""address_line1 is the primary address line with maximum 255 characters"""
    address_line2: NotRequired[str]
    r"""address_line2 is the secondary address line with maximum 255 characters"""
    address_postal_code: NotRequired[str]
    r"""address_postal_code is the ZIP code or postal code with maximum 20 characters"""
    address_state: NotRequired[str]
    r"""address_state is the state, province, or region name with maximum 100 characters"""
    email: NotRequired[str]
    r"""email is the customer's email address and must be a valid email format if provided"""
    integration_entity_mapping: NotRequired[List[DtoIntegrationEntityMappingTypedDict]]
    r"""integration_entity_mapping contains provider integration mappings for this customer"""
    metadata: NotRequired[Dict[str, str]]
    r"""metadata contains additional key-value pairs for storing extra information"""
    name: NotRequired[str]
    r"""name is the full name or company name of the customer"""
    parent_customer_external_id: NotRequired[str]
    r"""parent_customer_external_id is the external ID of the parent customer from your system
    Exactly one of parent_customer_id or parent_customer_external_id may be provided
    """
    parent_customer_id: NotRequired[str]
    r"""parent_customer_id is the internal FlexPrice ID of the parent customer"""
    skip_onboarding_workflow: NotRequired[bool]
    r"""skip_onboarding_workflow when true, prevents the customer onboarding workflow from being triggered
    This is used internally when a customer is created via a workflow to prevent infinite loops
    Default: false
    """
    tax_rate_overrides: NotRequired[List[DtoTaxRateOverrideTypedDict]]
    r"""tax_rate_overrides contains tax rate configurations to be linked to this customer"""


class DtoCreateCustomerRequest(BaseModel):
    r"""Request object for creating a new customer in the system"""

    external_id: str
    r"""external_id is the unique identifier from your system to reference this customer (required)"""

    address_city: Optional[str] = None
    r"""address_city is the city name with maximum 100 characters"""

    address_country: Optional[str] = None
    r"""address_country is the two-letter ISO 3166-1 alpha-2 country code"""

    address_line1: Optional[str] = None
    r"""address_line1 is the primary address line with maximum 255 characters"""

    address_line2: Optional[str] = None
    r"""address_line2 is the secondary address line with maximum 255 characters"""

    address_postal_code: Optional[str] = None
    r"""address_postal_code is the ZIP code or postal code with maximum 20 characters"""

    address_state: Optional[str] = None
    r"""address_state is the state, province, or region name with maximum 100 characters"""

    email: Optional[str] = None
    r"""email is the customer's email address and must be a valid email format if provided"""

    integration_entity_mapping: Optional[List[DtoIntegrationEntityMapping]] = None
    r"""integration_entity_mapping contains provider integration mappings for this customer"""

    metadata: Optional[Dict[str, str]] = None
    r"""metadata contains additional key-value pairs for storing extra information"""

    name: Optional[str] = None
    r"""name is the full name or company name of the customer"""

    parent_customer_external_id: Optional[str] = None
    r"""parent_customer_external_id is the external ID of the parent customer from your system
    Exactly one of parent_customer_id or parent_customer_external_id may be provided
    """

    parent_customer_id: Optional[str] = None
    r"""parent_customer_id is the internal FlexPrice ID of the parent customer"""

    skip_onboarding_workflow: Optional[bool] = None
    r"""skip_onboarding_workflow when true, prevents the customer onboarding workflow from being triggered
    This is used internally when a customer is created via a workflow to prevent infinite loops
    Default: false
    """

    tax_rate_overrides: Optional[List[DtoTaxRateOverride]] = None
    r"""tax_rate_overrides contains tax rate configurations to be linked to this customer"""

    @model_serializer(mode="wrap")
    def serialize_model(self, handler):
        optional_fields = set(
            [
                "address_city",
                "address_country",
                "address_line1",
                "address_line2",
                "address_postal_code",
                "address_state",
                "email",
                "integration_entity_mapping",
                "metadata",
                "name",
                "parent_customer_external_id",
                "parent_customer_id",
                "skip_onboarding_workflow",
                "tax_rate_overrides",
            ]
        )
        serialized = handler(self)
        m = {}

        for n, f in type(self).model_fields.items():
            k = f.alias or n
            val = serialized.get(k, serialized.get(n))

            if val != UNSET_SENTINEL:
                if val is not None or k not in optional_fields:
                    m[k] = val

        return m



"""Code generated by Speakeasy (https://speakeasy.com). DO NOT EDIT."""

from __future__ import annotations
from .aggregationtype import AggregationType
from .windowsize import WindowSize
from flexprice.types import BaseModel, UNSET_SENTINEL
from pydantic import model_serializer
from typing import Optional
from typing_extensions import NotRequired, TypedDict


class MeterAggregationTypedDict(TypedDict):
    bucket_size: NotRequired[WindowSize]
    expression: NotRequired[str]
    r"""Expression is an optional CEL expression to compute per-event quantity from event.properties.
    When set, it replaces Field-based extraction. Property names are used directly (e.g., token * duration * pixel).
    """
    field: NotRequired[str]
    r"""Field is the key in $event.properties on which the aggregation is to be applied
    For ex if the aggregation type is sum for API usage, the field could be \"duration_ms\" 
    Ignored when Expression is set.
    """
    group_by: NotRequired[str]
    r"""GroupBy is the property name in event.properties to group by before aggregating.
    Currently only supported for MAX aggregation with bucket_size.
    When set, aggregation is applied per unique value of this property within each bucket,
    then the per-group results are summed to produce the bucket total.
    """
    multiplier: NotRequired[str]
    r"""Multiplier is the multiplier for the aggregation
    For ex if the aggregation type is sum_with_multiplier for API usage, the multiplier could be 1000
    to scale up by a factor of 1000. If not provided, it will be null.
    """
    type: NotRequired[AggregationType]


class MeterAggregation(BaseModel):
    bucket_size: Optional[WindowSize] = None

    expression: Optional[str] = None
    r"""Expression is an optional CEL expression to compute per-event quantity from event.properties.
    When set, it replaces Field-based extraction. Property names are used directly (e.g., token * duration * pixel).
    """

    field: Optional[str] = None
    r"""Field is the key in $event.properties on which the aggregation is to be applied
    For ex if the aggregation type is sum for API usage, the field could be \"duration_ms\" 
    Ignored when Expression is set.
    """

    group_by: Optional[str] = None
    r"""GroupBy is the property name in event.properties to group by before aggregating.
    Currently only supported for MAX aggregation with bucket_size.
    When set, aggregation is applied per unique value of this property within each bucket,
    then the per-group results are summed to produce the bucket total.
    """

    multiplier: Optional[str] = None
    r"""Multiplier is the multiplier for the aggregation
    For ex if the aggregation type is sum_with_multiplier for API usage, the multiplier could be 1000
    to scale up by a factor of 1000. If not provided, it will be null.
    """

    type: Optional[AggregationType] = None

    @model_serializer(mode="wrap")
    def serialize_model(self, handler):
        optional_fields = set(
            ["bucket_size", "expression", "field", "group_by", "multiplier", "type"]
        )
        serialized = handler(self)
        m = {}

        for n, f in type(self).model_fields.items():
            k = f.alias or n
            val = serialized.get(k, serialized.get(n))

            if val != UNSET_SENTINEL:
                if val is not None or k not in optional_fields:
                    m[k] = val

        return m



"""Code generated by Speakeasy (https://speakeasy.com). DO NOT EDIT."""

from __future__ import annotations
from .datatype import DataType
from .filteroperatortype import FilterOperatorType
from .value import Value, ValueTypedDict
from flexprice.types import BaseModel, UNSET_SENTINEL
from pydantic import model_serializer
from typing import Optional
from typing_extensions import NotRequired, TypedDict


class FilterConditionTypedDict(TypedDict):
    data_type: NotRequired[DataType]
    field: NotRequired[str]
    operator: NotRequired[FilterOperatorType]
    value: NotRequired[ValueTypedDict]


class FilterCondition(BaseModel):
    data_type: Optional[DataType] = None

    field: Optional[str] = None

    operator: Optional[FilterOperatorType] = None

    value: Optional[Value] = None

    @model_serializer(mode="wrap")
    def serialize_model(self, handler):
        optional_fields = set(["data_type", "field", "operator", "value"])
        serialized = handler(self)
        m = {}

        for n, f in type(self).model_fields.items():
            k = f.alias or n
            val = serialized.get(k, serialized.get(n))

            if val != UNSET_SENTINEL:
                if val is not None or k not in optional_fields:
                    m[k] = val

        return m



"""Code generated by Speakeasy (https://speakeasy.com). DO NOT EDIT."""

from __future__ import annotations
from .dto_couponapplicationresponse import (
    DtoCouponApplicationResponse,
    DtoCouponApplicationResponseTypedDict,
)
from .dto_customerresponse import DtoCustomerResponse, DtoCustomerResponseTypedDict
from .dto_invoicelineitemresponse import (
    DtoInvoiceLineItemResponse,
    DtoInvoiceLineItemResponseTypedDict,
)
from .dto_taxappliedresponse import (
    DtoTaxAppliedResponse,
    DtoTaxAppliedResponseTypedDict,
)
from .invoicestatus import InvoiceStatus
from .invoicetype import InvoiceType
from .paymentstatus import PaymentStatus
from .status import Status
from flexprice.types import BaseModel, UNSET_SENTINEL
from pydantic import model_serializer
from typing import Dict, List, Optional, TYPE_CHECKING
from typing_extensions import NotRequired, TypedDict

if TYPE_CHECKING:
    from .dto_subscriptionresponse import (
        DtoSubscriptionResponse,
        DtoSubscriptionResponseTypedDict,
    )


class DtoInvoiceResponseTypedDict(TypedDict):
    adjustment_amount: NotRequired[str]
    r"""adjustment_amount is the total sum of credit notes of type \"adjustment\".
    These are non-cash reductions applied to the invoice (e.g. goodwill credit, billing correction).
    """
    amount_due: NotRequired[str]
    r"""amount_due is the total amount that needs to be paid for this invoice"""
    amount_paid: NotRequired[str]
    r"""amount_paid is the amount that has already been paid towards this invoice"""
    amount_remaining: NotRequired[str]
    r"""amount_remaining is the outstanding amount still owed on this invoice (calculated as amount_due minus amount_paid)"""
    billing_period: NotRequired[str]
    r"""billing_period describes the billing period this invoice covers (e.g., \"January 2024\", \"Q1 2024\")"""
    billing_reason: NotRequired[str]
    r"""billing_reason indicates why this invoice was generated (e.g., \"subscription_billing\", \"manual_charge\")"""
    billing_sequence: NotRequired[int]
    r"""billing_sequence is the sequential number indicating the billing cycle for subscription invoices"""
    coupon_applications: NotRequired[List[DtoCouponApplicationResponseTypedDict]]
    r"""coupon_applications contains the coupon applications associated with this invoice (overrides embedded field)"""
    created_at: NotRequired[str]
    created_by: NotRequired[str]
    currency: NotRequired[str]
    r"""currency is the three-letter ISO currency code (e.g., USD, EUR, GBP) that applies to all monetary amounts on this invoice"""
    customer: NotRequired[DtoCustomerResponseTypedDict]
    r"""Customer response object containing all customer information"""
    customer_id: NotRequired[str]
    r"""customer_id is the ID of the customer who will receive this invoice"""
    description: NotRequired[str]
    r"""description is an optional description or notes about this invoice"""
    due_date: NotRequired[str]
    r"""due_date is the date when payment for this invoice is due"""
    environment_id: NotRequired[str]
    r"""environment_id is the ID of the environment this invoice belongs to (for multi-environment setups)"""
    finalized_at: NotRequired[str]
    r"""finalized_at is the timestamp when this invoice was finalized and made ready for payment"""
    id: NotRequired[str]
    r"""id is the unique identifier for this invoice"""
    idempotency_key: NotRequired[str]
    r"""idempotency_key is a unique key used to prevent duplicate invoice creation when retrying API calls"""
    invoice_number: NotRequired[str]
    r"""invoice_number is the human-readable invoice number displayed to customers (e.g., INV-2024-001)"""
    invoice_pdf_url: NotRequired[str]
    r"""invoice_pdf_url is the URL where customers can download the PDF version of this invoice"""
    invoice_status: NotRequired[InvoiceStatus]
    invoice_type: NotRequired[InvoiceType]
    line_items: NotRequired[List[DtoInvoiceLineItemResponseTypedDict]]
    r"""line_items contains the individual items that make up this invoice (overrides embedded field)"""
    metadata: NotRequired[Dict[str, str]]
    overpaid_amount: NotRequired[str]
    r"""overpaid_amount is the amount overpaid if payment_status is OVERPAID (amount_paid - total)"""
    paid_at: NotRequired[str]
    r"""paid_at is the timestamp when this invoice was fully paid"""
    payment_status: NotRequired[PaymentStatus]
    period_end: NotRequired[str]
    r"""period_end is the end date of the billing period covered by this invoice"""
    period_start: NotRequired[str]
    r"""period_start is the start date of the billing period covered by this invoice"""
    recalculated_invoice_id: NotRequired[str]
    r"""recalculated_invoice_id is the ID of the replacement invoice created when this invoice was voided and recalculated.
    When set, it forms a parent→child link from this (voided) invoice to the new replacement invoice.
    """
    refunded_amount: NotRequired[str]
    r"""refunded_amount is the total sum of credit notes of type \"refund\".
    These are actual refunds issued to the customer.
    """
    status: NotRequired[Status]
    subscription: NotRequired["DtoSubscriptionResponseTypedDict"]
    subscription_id: NotRequired[str]
    r"""subscription_id is the ID of the subscription this invoice is associated with (only present for subscription-based invoices)"""
    subtotal: NotRequired[str]
    r"""subtotal is the sum of all line items before any taxes, discounts, or additional fees"""
    taxes: NotRequired[List[DtoTaxAppliedResponseTypedDict]]
    r"""tax_applied_records contains the tax applied records associated with this invoice"""
    tenant_id: NotRequired[str]
    total: NotRequired[str]
    r"""total is the final amount including taxes, fees, and discounts"""
    total_discount: NotRequired[str]
    r"""total_discount is the sum of all coupon discounts applied to the invoice"""
    total_prepaid_credits_applied: NotRequired[str]
    r"""total_prepaid_credits_applied is the total amount of prepaid credits applied to this invoice."""
    total_tax: NotRequired[str]
    r"""total_tax is the sum of all taxes combined at the invoice level."""
    updated_at: NotRequired[str]
    updated_by: NotRequired[str]
    version: NotRequired[int]
    r"""version is the version number for tracking changes to this invoice"""
    voided_at: NotRequired[str]
    r"""voided_at is the timestamp when this invoice was voided or cancelled"""


class DtoInvoiceResponse(BaseModel):
    adjustment_amount: Optional[str] = None
    r"""adjustment_amount is the total sum of credit notes of type \"adjustment\".
    These are non-cash reductions applied to the invoice (e.g. goodwill credit, billing correction).
    """

    amount_due: Optional[str] = None
    r"""amount_due is the total amount that needs to be paid for this invoice"""

    amount_paid: Optional[str] = None
    r"""amount_paid is the amount that has already been paid towards this invoice"""

    amount_remaining: Optional[str] = None
    r"""amount_remaining is the outstanding amount still owed on this invoice (calculated as amount_due minus amount_paid)"""

    billing_period: Optional[str] = None
    r"""billing_period describes the billing period this invoice covers (e.g., \"January 2024\", \"Q1 2024\")"""

    billing_reason: Optional[str] = None
    r"""billing_reason indicates why this invoice was generated (e.g., \"subscription_billing\", \"manual_charge\")"""

    billing_sequence: Optional[int] = None
    r"""billing_sequence is the sequential number indicating the billing cycle for subscription invoices"""

    coupon_applications: Optional[List[DtoCouponApplicationResponse]] = None
    r"""coupon_applications contains the coupon applications associated with this invoice (overrides embedded field)"""

    created_at: Optional[str] = None

    created_by: Optional[str] = None

    currency: Optional[str] = None
    r"""currency is the three-letter ISO currency code (e.g., USD, EUR, GBP) that applies to all monetary amounts on this invoice"""

    customer: Optional[DtoCustomerResponse] = None
    r"""Customer response object containing all customer information"""

    customer_id: Optional[str] = None
    r"""customer_id is the ID of the customer who will receive this invoice"""

    description: Optional[str] = None
    r"""description is an optional description or notes about this invoice"""

    due_date: Optional[str] = None
    r"""due_date is the date when payment for this invoice is due"""

    environment_id: Optional[str] = None
    r"""environment_id is the ID of the environment this invoice belongs to (for multi-environment setups)"""

    finalized_at: Optional[str] = None
    r"""finalized_at is the timestamp when this invoice was finalized and made ready for payment"""

    id: Optional[str] = None
    r"""id is the unique identifier for this invoice"""

    idempotency_key: Optional[str] = None
    r"""idempotency_key is a unique key used to prevent duplicate invoice creation when retrying API calls"""

    invoice_number: Optional[str] = None
    r"""invoice_number is the human-readable invoice number displayed to customers (e.g., INV-2024-001)"""

    invoice_pdf_url: Optional[str] = None
    r"""invoice_pdf_url is the URL where customers can download the PDF version of this invoice"""

    invoice_status: Optional[InvoiceStatus] = None

    invoice_type: Optional[InvoiceType] = None

    line_items: Optional[List[DtoInvoiceLineItemResponse]] = None
    r"""line_items contains the individual items that make up this invoice (overrides embedded field)"""

    metadata: Optional[Dict[str, str]] = None

    overpaid_amount: Optional[str] = None
    r"""overpaid_amount is the amount overpaid if payment_status is OVERPAID (amount_paid - total)"""

    paid_at: Optional[str] = None
    r"""paid_at is the timestamp when this invoice was fully paid"""

    payment_status: Optional[PaymentStatus] = None

    period_end: Optional[str] = None
    r"""period_end is the end date of the billing period covered by this invoice"""

    period_start: Optional[str] = None
    r"""period_start is the start date of the billing period covered by this invoice"""

    recalculated_invoice_id: Optional[str] = None
    r"""recalculated_invoice_id is the ID of the replacement invoice created when this invoice was voided and recalculated.
    When set, it forms a parent→child link from this (voided) invoice to the new replacement invoice.
    """

    refunded_amount: Optional[str] = None
    r"""refunded_amount is the total sum of credit notes of type \"refund\".
    These are actual refunds issued to the customer.
    """

    status: Optional[Status] = None

    subscription: Optional["DtoSubscriptionResponse"] = None

    subscription_id: Optional[str] = None
    r"""subscription_id is the ID of the subscription this invoice is associated with (only present for subscription-based invoices)"""

    subtotal: Optional[str] = None
    r"""subtotal is the sum of all line items before any taxes, discounts, or additional fees"""

    taxes: Optional[List[DtoTaxAppliedResponse]] = None
    r"""tax_applied_records contains the tax applied records associated with this invoice"""

    tenant_id: Optional[str] = None

    total: Optional[str] = None
    r"""total is the final amount including taxes, fees, and discounts"""

    total_discount: Optional[str] = None
    r"""total_discount is the sum of all coupon discounts applied to the invoice"""

    total_prepaid_credits_applied: Optional[str] = None
    r"""total_prepaid_credits_applied is the total amount of prepaid credits applied to this invoice."""

    total_tax: Optional[str] = None
    r"""total_tax is the sum of all taxes combined at the invoice level."""

    updated_at: Optional[str] = None

    updated_by: Optional[str] = None

    version: Optional[int] = None
    r"""version is the version number for tracking changes to this invoice"""

    voided_at: Optional[str] = None
    r"""voided_at is the timestamp when this invoice was voided or cancelled"""

    @model_serializer(mode="wrap")
    def serialize_model(self, handler):
        optional_fields = set(
            [
                "adjustment_amount",
                "amount_due",
                "amount_paid",
                "amount_remaining",
                "billing_period",
                "billing_reason",
                "billing_sequence",
                "coupon_applications",
                "created_at",
                "created_by",
                "currency",
                "customer",
                "customer_id",
                "description",
                "due_date",
                "environment_id",
                "finalized_at",
                "id",
                "idempotency_key",
                "invoice_number",
                "invoice_pdf_url",
                "invoice_status",
                "invoice_type",
                "line_items",
                "metadata",
                "overpaid_amount",
                "paid_at",
                "payment_status",
                "period_end",
                "period_start",
                "recalculated_invoice_id",
                "refunded_amount",
                "status",
                "subscription",
                "subscription_id",
                "subtotal",
                "taxes",
                "tenant_id",
                "total",
                "total_discount",
                "total_prepaid_credits_applied",
                "total_tax",
                "updated_at",
                "updated_by",
                "version",
                "voided_at",
            ]
        )
        serialized = handler(self)
        m = {}

        for n, f in type(self).model_fields.items():
            k = f.alias or n
            val = serialized.get(k, serialized.get(n))

            if val != UNSET_SENTINEL:
                if val is not None or k not in optional_fields:
                    m[k] = val

        return m



"""Code generated by Speakeasy (https://speakeasy.com). DO NOT EDIT."""

from __future__ import annotations
from .dto_costanalyticitem import DtoCostAnalyticItem, DtoCostAnalyticItemTypedDict
from flexprice.types import BaseModel, UNSET_SENTINEL
from pydantic import model_serializer
from typing import List, Optional
from typing_extensions import NotRequired, TypedDict


class DtoGetDetailedCostAnalyticsResponseTypedDict(TypedDict):
    cost_analytics: NotRequired[List[DtoCostAnalyticItemTypedDict]]
    r"""Cost analytics array (flattened from nested structure)"""
    currency: NotRequired[str]
    end_time: NotRequired[str]
    margin: NotRequired[str]
    r"""Revenue - Cost"""
    margin_percent: NotRequired[str]
    r"""(Margin / Revenue) * 100"""
    roi: NotRequired[str]
    r"""(Revenue - Cost) / Cost"""
    roi_percent: NotRequired[str]
    r"""ROI * 100"""
    start_time: NotRequired[str]
    total_cost: NotRequired[str]
    total_revenue: NotRequired[str]
    r"""Derived metrics"""


class DtoGetDetailedCostAnalyticsResponse(BaseModel):
    cost_analytics: Optional[List[DtoCostAnalyticItem]] = None
    r"""Cost analytics array (flattened from nested structure)"""

    currency: Optional[str] = None

    end_time: Optional[str] = None

    margin: Optional[str] = None
    r"""Revenue - Cost"""

    margin_percent: Optional[str] = None
    r"""(Margin / Revenue) * 100"""

    roi: Optional[str] = None
    r"""(Revenue - Cost) / Cost"""

    roi_percent: Optional[str] = None
    r"""ROI * 100"""

    start_time: Optional[str] = None

    total_cost: Optional[str] = None

    total_revenue: Optional[str] = None
    r"""Derived metrics"""

    @model_serializer(mode="wrap")
    def serialize_model(self, handler):
        optional_fields = set(
            [
                "cost_analytics",
                "currency",
                "end_time",
                "margin",
                "margin_percent",
                "roi",
                "roi_percent",
                "start_time",
                "total_cost",
                "total_revenue",
            ]
        )
        serialized = handler(self)
        m = {}

        for n, f in type(self).model_fields.items():
            k = f.alias or n
            val = serialized.get(k, serialized.get(n))

            if val != UNSET_SENTINEL:
                if val is not None or k not in optional_fields:
                    m[k] = val

        return m



"""Code generated by Speakeasy (https://speakeasy.com). DO NOT EDIT."""

from __future__ import annotations
from flexprice.types import UnrecognizedStr
from typing import Literal, Union


TaxRateStatus = Union[
    Literal[
        "ACTIVE",
        "INACTIVE",
    ],
    UnrecognizedStr,
]



"""Code generated by Speakeasy (https://speakeasy.com). DO NOT EDIT."""

from __future__ import annotations
from .filtercondition import FilterCondition, FilterConditionTypedDict
from .invoicestatus import InvoiceStatus
from .invoicetype import InvoiceType
from .paymentstatus import PaymentStatus
from .sortcondition import SortCondition, SortConditionTypedDict
from .status import Status
from flexprice.types import BaseModel, UNSET_SENTINEL
from pydantic import model_serializer
from typing import List, Literal, Optional
from typing_extensions import NotRequired, TypedDict


InvoiceFilterOrder = Literal[
    "asc",
    "desc",
]


class InvoiceFilterTypedDict(TypedDict):
    amount_due_gt: NotRequired[float]
    r"""amount_due_gt filters invoices with a total amount due greater than the specified value
    Useful for finding invoices above a certain threshold or identifying high-value invoices
    """
    amount_remaining_gt: NotRequired[float]
    r"""amount_remaining_gt filters invoices with an outstanding balance greater than the specified value
    Useful for finding invoices that still have significant unpaid amounts
    """
    customer_id: NotRequired[str]
    r"""customer_id filters invoices for a specific customer using FlexPrice's internal customer ID
    This is the ID returned by FlexPrice when creating or retrieving customers
    """
    end_time: NotRequired[str]
    expand: NotRequired[str]
    external_customer_id: NotRequired[str]
    r"""external_customer_id filters invoices for a customer using your system's customer identifier
    This is the ID you provided when creating the customer in FlexPrice
    """
    filters: NotRequired[List[FilterConditionTypedDict]]
    invoice_ids: NotRequired[List[str]]
    r"""invoice_ids restricts results to invoices with the specified IDs
    Use this to retrieve specific invoices when you know their exact identifiers
    """
    invoice_status: NotRequired[List[InvoiceStatus]]
    r"""invoice_status filters by the current state of invoices in their lifecycle
    Multiple statuses can be specified to include invoices in any of the listed states
    """
    invoice_type: NotRequired[InvoiceType]
    limit: NotRequired[int]
    offset: NotRequired[int]
    order: NotRequired[InvoiceFilterOrder]
    payment_status: NotRequired[List[PaymentStatus]]
    r"""payment_status filters by the payment state of invoices
    Multiple statuses can be specified to include invoices with any of the listed payment states
    """
    period_end_gte: NotRequired[str]
    r"""period_end_gte filters invoices with period_end >= value"""
    period_end_lte: NotRequired[str]
    r"""period_end_lte filters invoices with period_end <= value"""
    period_start_gte: NotRequired[str]
    r"""period_start_gte filters invoices with period_start >= value"""
    period_start_lte: NotRequired[str]
    r"""period_start_lte filters invoices with period_start <= value"""
    skip_line_items: NotRequired[bool]
    r"""SkipLineItems if true, will not include line items in the response"""
    sort: NotRequired[List[SortConditionTypedDict]]
    start_time: NotRequired[str]
    status: NotRequired[Status]
    subscription_id: NotRequired[str]
    r"""subscription_id filters invoices generated for a specific subscription
    Only returns invoices that were created as part of the specified subscription's billing
    """


class InvoiceFilter(BaseModel):
    amount_due_gt: Optional[float] = None
    r"""amount_due_gt filters invoices with a total amount due greater than the specified value
    Useful for finding invoices above a certain threshold or identifying high-value invoices
    """

    amount_remaining_gt: Optional[float] = None
    r"""amount_remaining_gt filters invoices with an outstanding balance greater than the specified value
    Useful for finding invoices that still have significant unpaid amounts
    """

    customer_id: Optional[str] = None
    r"""customer_id filters invoices for a specific customer using FlexPrice's internal customer ID
    This is the ID returned by FlexPrice when creating or retrieving customers
    """

    end_time: Optional[str] = None

    expand: Optional[str] = None

    external_customer_id: Optional[str] = None
    r"""external_customer_id filters invoices for a customer using your system's customer identifier
    This is the ID you provided when creating the customer in FlexPrice
    """

    filters: Optional[List[FilterCondition]] = None

    invoice_ids: Optional[List[str]] = None
    r"""invoice_ids restricts results to invoices with the specified IDs
    Use this to retrieve specific invoices when you know their exact identifiers
    """

    invoice_status: Optional[List[InvoiceStatus]] = None
    r"""invoice_status filters by the current state of invoices in their lifecycle
    Multiple statuses can be specified to include invoices in any of the listed states
    """

    invoice_type: Optional[InvoiceType] = None

    limit: Optional[int] = None

    offset: Optional[int] = None

    order: Optional[InvoiceFilterOrder] = None

    payment_status: Optional[List[PaymentStatus]] = None
    r"""payment_status filters by the payment state of invoices
    Multiple statuses can be specified to include invoices with any of the listed payment states
    """

    period_end_gte: Optional[str] = None
    r"""period_end_gte filters invoices with period_end >= value"""

    period_end_lte: Optional[str] = None
    r"""period_end_lte filters invoices with period_end <= value"""

    period_start_gte: Optional[str] = None
    r"""period_start_gte filters invoices with period_start >= value"""

    period_start_lte: Optional[str] = None
    r"""period_start_lte filters invoices with period_start <= value"""

    skip_line_items: Optional[bool] = None
    r"""SkipLineItems if true, will not include line items in the response"""

    sort: Optional[List[SortCondition]] = None

    start_time: Optional[str] = None

    status: Optional[Status] = None

    subscription_id: Optional[str] = None
    r"""subscription_id filters invoices generated for a specific subscription
    Only returns invoices that were created as part of the specified subscription's billing
    """

    @model_serializer(mode="wrap")
    def serialize_model(self, handler):
        optional_fields = set(
            [
                "amount_due_gt",
                "amount_remaining_gt",
                "customer_id",
                "end_time",
                "expand",
                "external_customer_id",
                "filters",
                "invoice_ids",
                "invoice_status",
                "invoice_type",
                "limit",
                "offset",
                "order",
                "payment_status",
                "period_end_gte",
                "period_end_lte",
                "period_start_gte",
                "period_start_lte",
                "skip_line_items",
                "sort",
                "start_time",
                "status",
                "subscription_id",
            ]
        )
        serialized = handler(self)
        m = {}

        for n, f in type(self).model_fields.items():
            k = f.alias or n
            val = serialized.get(k, serialized.get(n))

            if val != UNSET_SENTINEL:
                if val is not None or k not in optional_fields:
                    m[k] = val

        return m



"""Code generated by Speakeasy (https://speakeasy.com). DO NOT EDIT."""

from __future__ import annotations
from dataclasses import dataclass, field
from flexprice.models import errors_errordetail as models_errors_errordetail
from flexprice.models.errors import FlexpriceError
from flexprice.types import BaseModel
import httpx
from typing import Optional


class ErrorsErrorResponseData(BaseModel):
    error: Optional[models_errors_errordetail.ErrorsErrorDetail] = None
    success: Optional[bool] = None


@dataclass(unsafe_hash=True)
class ErrorsErrorResponse(FlexpriceError):
    data: ErrorsErrorResponseData = field(hash=False)

    def __init__(
        self,
        data: ErrorsErrorResponseData,
        raw_response: httpx.Response,
        body: Optional[str] = None,
    ):
        fallback = body or raw_response.text
        message = str(data.error.message) if data.error else fallback
        super().__init__(message, raw_response, body)
        object.__setattr__(self, "data", data)



"""Code generated by Speakeasy (https://speakeasy.com). DO NOT EDIT."""

from __future__ import annotations
from .dto_integrationentitymapping import (
    DtoIntegrationEntityMapping,
    DtoIntegrationEntityMappingTypedDict,
)
from flexprice.types import BaseModel, UNSET_SENTINEL
from pydantic import model_serializer
from typing import Dict, List, Optional
from typing_extensions import NotRequired, TypedDict


class DtoUpdateCustomerRequestTypedDict(TypedDict):
    r"""Request object for updating an existing customer. All fields are optional - only provided fields will be updated"""

    address_city: NotRequired[str]
    r"""address_city is the updated city name with maximum 100 characters"""
    address_country: NotRequired[str]
    r"""address_country is the updated two-letter ISO 3166-1 alpha-2 country code"""
    address_line1: NotRequired[str]
    r"""address_line1 is the updated primary address line with maximum 255 characters"""
    address_line2: NotRequired[str]
    r"""address_line2 is the updated secondary address line with maximum 255 characters"""
    address_postal_code: NotRequired[str]
    r"""address_postal_code is the updated postal code with maximum 20 characters"""
    address_state: NotRequired[str]
    r"""address_state is the updated state, province, or region name with maximum 100 characters"""
    email: NotRequired[str]
    r"""email is the updated email address and must be a valid email format if provided"""
    external_id: NotRequired[str]
    r"""external_id is the updated external identifier for the customer"""
    integration_entity_mapping: NotRequired[List[DtoIntegrationEntityMappingTypedDict]]
    r"""integration_entity_mapping contains provider integration mappings for this customer"""
    metadata: NotRequired[Dict[str, str]]
    r"""metadata contains updated key-value pairs that will replace existing metadata"""
    name: NotRequired[str]
    r"""name is the updated name or company name for the customer"""
    parent_customer_external_id: NotRequired[str]
    r"""parent_customer_external_id is the external ID of the parent customer from your system
    Exactly one of parent_customer_id or parent_customer_external_id may be provided
    If you provide the external ID, the parent customer value will be ignored
    """
    parent_customer_id: NotRequired[str]
    r"""parent_customer_id is the internal FlexPrice ID of the parent customer"""


class DtoUpdateCustomerRequest(BaseModel):
    r"""Request object for updating an existing customer. All fields are optional - only provided fields will be updated"""

    address_city: Optional[str] = None
    r"""address_city is the updated city name with maximum 100 characters"""

    address_country: Optional[str] = None
    r"""address_country is the updated two-letter ISO 3166-1 alpha-2 country code"""

    address_line1: Optional[str] = None
    r"""address_line1 is the updated primary address line with maximum 255 characters"""

    address_line2: Optional[str] = None
    r"""address_line2 is the updated secondary address line with maximum 255 characters"""

    address_postal_code: Optional[str] = None
    r"""address_postal_code is the updated postal code with maximum 20 characters"""

    address_state: Optional[str] = None
    r"""address_state is the updated state, province, or region name with maximum 100 characters"""

    email: Optional[str] = None
    r"""email is the updated email address and must be a valid email format if provided"""

    external_id: Optional[str] = None
    r"""external_id is the updated external identifier for the customer"""

    integration_entity_mapping: Optional[List[DtoIntegrationEntityMapping]] = None
    r"""integration_entity_mapping contains provider integration mappings for this customer"""

    metadata: Optional[Dict[str, str]] = None
    r"""metadata contains updated key-value pairs that will replace existing metadata"""

    name: Optional[str] = None
    r"""name is the updated name or company name for the customer"""

    parent_customer_external_id: Optional[str] = None
    r"""parent_customer_external_id is the external ID of the parent customer from your system
    Exactly one of parent_customer_id or parent_customer_external_id may be provided
    If you provide the external ID, the parent customer value will be ignored
    """

    parent_customer_id: Optional[str] = None
    r"""parent_customer_id is the internal FlexPrice ID of the parent customer"""

    @model_serializer(mode="wrap")
    def serialize_model(self, handler):
        optional_fields = set(
            [
                "address_city",
                "address_country",
                "address_line1",
                "address_line2",
                "address_postal_code",
                "address_state",
                "email",
                "external_id",
                "integration_entity_mapping",
                "metadata",
                "name",
                "parent_customer_external_id",
                "parent_customer_id",
            ]
        )
        serialized = handler(self)
        m = {}

        for n, f in type(self).model_fields.items():
            k = f.alias or n
            val = serialized.get(k, serialized.get(n))

            if val != UNSET_SENTINEL:
                if val is not None or k not in optional_fields:
                    m[k] = val

        return m



"""Code generated by Speakeasy (https://speakeasy.com). DO NOT EDIT."""

from __future__ import annotations
from .s3compressiontype import S3CompressionType
from .s3encryptiontype import S3EncryptionType
from flexprice.types import BaseModel, UNSET_SENTINEL
from pydantic import model_serializer
from typing import Optional
from typing_extensions import NotRequired, TypedDict


class S3JobConfigTypedDict(TypedDict):
    bucket: NotRequired[str]
    r"""S3 bucket name"""
    compression: NotRequired[S3CompressionType]
    encryption: NotRequired[S3EncryptionType]
    endpoint_url: NotRequired[str]
    r"""Custom S3 endpoint URL (e.g., \"http://minio:9000\" for MinIO)"""
    key_prefix: NotRequired[str]
    r"""Optional prefix for S3 keys (e.g., \"flexprice-exports/\")"""
    region: NotRequired[str]
    r"""AWS region (e.g., \"us-west-2\")"""
    use_path_style: NotRequired[bool]
    r"""Use path-style addressing instead of virtual-hosted-style (required for MinIO)"""


class S3JobConfig(BaseModel):
    bucket: Optional[str] = None
    r"""S3 bucket name"""

    compression: Optional[S3CompressionType] = None

    encryption: Optional[S3EncryptionType] = None

    endpoint_url: Optional[str] = None
    r"""Custom S3 endpoint URL (e.g., \"http://minio:9000\" for MinIO)"""

    key_prefix: Optional[str] = None
    r"""Optional prefix for S3 keys (e.g., \"flexprice-exports/\")"""

    region: Optional[str] = None
    r"""AWS region (e.g., \"us-west-2\")"""

    use_path_style: Optional[bool] = None
    r"""Use path-style addressing instead of virtual-hosted-style (required for MinIO)"""

    @model_serializer(mode="wrap")
    def serialize_model(self, handler):
        optional_fields = set(
            [
                "bucket",
                "compression",
                "encryption",
                "endpoint_url",
                "key_prefix",
                "region",
                "use_path_style",
            ]
        )
        serialized = handler(self)
        m = {}

        for n, f in type(self).model_fields.items():
            k = f.alias or n
            val = serialized.get(k, serialized.get(n))

            if val != UNSET_SENTINEL:
                if val is not None or k not in optional_fields:
                    m[k] = val

        return m



"""Code generated by Speakeasy (https://speakeasy.com). DO NOT EDIT."""

from __future__ import annotations
from .applicationstatus import ApplicationStatus
from .creditgrantapplicationreason import CreditGrantApplicationReason
from .status import Status
from .subscriptionstatus import SubscriptionStatus
from flexprice.types import BaseModel, UNSET_SENTINEL
from pydantic import model_serializer
from typing import Dict, Optional
from typing_extensions import NotRequired, TypedDict


class DtoCreditGrantApplicationResponseTypedDict(TypedDict):
    application_reason: NotRequired[CreditGrantApplicationReason]
    application_status: NotRequired[ApplicationStatus]
    applied_at: NotRequired[str]
    created_at: NotRequired[str]
    created_by: NotRequired[str]
    credit_grant_id: NotRequired[str]
    credits: NotRequired[float]
    environment_id: NotRequired[str]
    failure_reason: NotRequired[str]
    id: NotRequired[str]
    idempotency_key: NotRequired[str]
    metadata: NotRequired[Dict[str, str]]
    period_end: NotRequired[str]
    period_start: NotRequired[str]
    retry_count: NotRequired[int]
    scheduled_for: NotRequired[str]
    status: NotRequired[Status]
    subscription_id: NotRequired[str]
    subscription_status_at_application: NotRequired[SubscriptionStatus]
    tenant_id: NotRequired[str]
    updated_at: NotRequired[str]
    updated_by: NotRequired[str]


class DtoCreditGrantApplicationResponse(BaseModel):
    application_reason: Optional[CreditGrantApplicationReason] = None

    application_status: Optional[ApplicationStatus] = None

    applied_at: Optional[str] = None

    created_at: Optional[str] = None

    created_by: Optional[str] = None

    credit_grant_id: Optional[str] = None

    credits: Optional[float] = None

    environment_id: Optional[str] = None

    failure_reason: Optional[str] = None

    id: Optional[str] = None

    idempotency_key: Optional[str] = None

    metadata: Optional[Dict[str, str]] = None

    period_end: Optional[str] = None

    period_start: Optional[str] = None

    retry_count: Optional[int] = None

    scheduled_for: Optional[str] = None

    status: Optional[Status] = None

    subscription_id: Optional[str] = None

    subscription_status_at_application: Optional[SubscriptionStatus] = None

    tenant_id: Optional[str] = None

    updated_at: Optional[str] = None

    updated_by: Optional[str] = None

    @model_serializer(mode="wrap")
    def serialize_model(self, handler):
        optional_fields = set(
            [
                "application_reason",
                "application_status",
                "applied_at",
                "created_at",
                "created_by",
                "credit_grant_id",
                "credits",
                "environment_id",
                "failure_reason",
                "id",
                "idempotency_key",
                "metadata",
                "period_end",
                "period_start",
                "retry_count",
                "scheduled_for",
                "status",
                "subscription_id",
                "subscription_status_at_application",
                "tenant_id",
                "updated_at",
                "updated_by",
            ]
        )
        serialized = handler(self)
        m = {}

        for n, f in type(self).model_fields.items():
            k = f.alias or n
            val = serialized.get(k, serialized.get(n))

            if val != UNSET_SENTINEL:
                if val is not None or k not in optional_fields:
                    m[k] = val

        return m



"""Code generated by Speakeasy (https://speakeasy.com). DO NOT EDIT."""

from __future__ import annotations
from flexprice.types import UnrecognizedStr
from typing import Literal, Union


PaymentMethodType = Union[
    Literal[
        "CARD",
        "ACH",
        "OFFLINE",
        "CREDITS",
        "PAYMENT_LINK",
    ],
    UnrecognizedStr,
]



"""Code generated by Speakeasy (https://speakeasy.com). DO NOT EDIT."""

from __future__ import annotations
from .commitmentinfo import CommitmentInfo, CommitmentInfoTypedDict
from flexprice.types import BaseModel, UNSET_SENTINEL
from pydantic import model_serializer
from typing import Dict, Optional
from typing_extensions import NotRequired, TypedDict


class DtoCreateInvoiceLineItemRequestTypedDict(TypedDict):
    amount: str
    r"""amount is the monetary amount for this line item"""
    quantity: str
    r"""quantity is the quantity of units for this line item"""
    commitment_info: NotRequired[CommitmentInfoTypedDict]
    display_name: NotRequired[str]
    r"""display_name is the optional human-readable name for this line item"""
    entity_id: NotRequired[str]
    r"""entity_id is the optional unique identifier of the entity associated with this line item"""
    entity_type: NotRequired[str]
    r"""entity_type is the optional type of the entity associated with this line item"""
    invoice_level_discount: NotRequired[str]
    r"""invoice_level_discount is the discount amount in invoice currency applied to all line items on the invoice."""
    line_item_discount: NotRequired[str]
    r"""line_item_discount is the discount amount in invoice currency applied directly to this line item."""
    metadata: NotRequired[Dict[str, str]]
    meter_display_name: NotRequired[str]
    r"""meter_display_name is the optional human-readable name of the meter"""
    meter_id: NotRequired[str]
    r"""meter_id is the optional unique identifier of the meter used for usage tracking"""
    period_end: NotRequired[str]
    r"""period_end is the optional end date of the period this line item covers"""
    period_start: NotRequired[str]
    r"""period_start is the optional start date of the period this line item covers"""
    plan_display_name: NotRequired[str]
    r"""plan_display_name is the optional human-readable name of the plan"""
    plan_id: NotRequired[str]
    r"""TODO: !REMOVE after migration
    plan_id is the optional unique identifier of the plan associated with this line item
    """
    prepaid_credits_applied: NotRequired[str]
    r"""prepaid_credits_applied is the amount in invoice currency reduced from this line item due to prepaid credits application."""
    price_id: NotRequired[str]
    r"""price_id is the optional unique identifier of the price associated with this line item"""
    price_type: NotRequired[str]
    r"""price_type indicates the type of pricing (fixed, usage, tiered, etc.)"""
    price_unit: NotRequired[str]
    r"""price_unit is the optional 3-digit ISO code of the price unit associated with this line item"""
    price_unit_amount: NotRequired[str]
    r"""price_unit_amount is the optional amount converted to the price unit currency"""


class DtoCreateInvoiceLineItemRequest(BaseModel):
    amount: str
    r"""amount is the monetary amount for this line item"""

    quantity: str
    r"""quantity is the quantity of units for this line item"""

    commitment_info: Optional[CommitmentInfo] = None

    display_name: Optional[str] = None
    r"""display_name is the optional human-readable name for this line item"""

    entity_id: Optional[str] = None
    r"""entity_id is the optional unique identifier of the entity associated with this line item"""

    entity_type: Optional[str] = None
    r"""entity_type is the optional type of the entity associated with this line item"""

    invoice_level_discount: Optional[str] = None
    r"""invoice_level_discount is the discount amount in invoice currency applied to all line items on the invoice."""

    line_item_discount: Optional[str] = None
    r"""line_item_discount is the discount amount in invoice currency applied directly to this line item."""

    metadata: Optional[Dict[str, str]] = None

    meter_display_name: Optional[str] = None
    r"""meter_display_name is the optional human-readable name of the meter"""

    meter_id: Optional[str] = None
    r"""meter_id is the optional unique identifier of the meter used for usage tracking"""

    period_end: Optional[str] = None
    r"""period_end is the optional end date of the period this line item covers"""

    period_start: Optional[str] = None
    r"""period_start is the optional start date of the period this line item covers"""

    plan_display_name: Optional[str] = None
    r"""plan_display_name is the optional human-readable name of the plan"""

    plan_id: Optional[str] = None
    r"""TODO: !REMOVE after migration
    plan_id is the optional unique identifier of the plan associated with this line item
    """

    prepaid_credits_applied: Optional[str] = None
    r"""prepaid_credits_applied is the amount in invoice currency reduced from this line item due to prepaid credits application."""

    price_id: Optional[str] = None
    r"""price_id is the optional unique identifier of the price associated with this line item"""

    price_type: Optional[str] = None
    r"""price_type indicates the type of pricing (fixed, usage, tiered, etc.)"""

    price_unit: Optional[str] = None
    r"""price_unit is the optional 3-digit ISO code of the price unit associated with this line item"""

    price_unit_amount: Optional[str] = None
    r"""price_unit_amount is the optional amount converted to the price unit currency"""

    @model_serializer(mode="wrap")
    def serialize_model(self, handler):
        optional_fields = set(
            [
                "commitment_info",
                "display_name",
                "entity_id",
                "entity_type",
                "invoice_level_discount",
                "line_item_discount",
                "metadata",
                "meter_display_name",
                "meter_id",
                "period_end",
                "period_start",
                "plan_display_name",
                "plan_id",
                "prepaid_credits_applied",
                "price_id",
                "price_type",
                "price_unit",
                "price_unit_amount",
            ]
        )
        serialized = handler(self)
        m = {}

        for n, f in type(self).model_fields.items():
            k = f.alias or n
            val = serialized.get(k, serialized.get(n))

            if val != UNSET_SENTINEL:
                if val is not None or k not in optional_fields:
                    m[k] = val

        return m



"""Code generated by Speakeasy (https://speakeasy.com). DO NOT EDIT."""

from __future__ import annotations
from .dto_ingesteventrequest import (
    DtoIngestEventRequest,
    DtoIngestEventRequestTypedDict,
)
from flexprice.types import BaseModel
from typing import List
from typing_extensions import TypedDict


class DtoBulkIngestEventRequestTypedDict(TypedDict):
    events: List[DtoIngestEventRequestTypedDict]


class DtoBulkIngestEventRequest(BaseModel):
    events: List[DtoIngestEventRequest]



"""Code generated by Speakeasy (https://speakeasy.com). DO NOT EDIT."""

from __future__ import annotations
from .subscription_subscriptionlineitem import (
    SubscriptionSubscriptionLineItem,
    SubscriptionSubscriptionLineItemTypedDict,
)
from flexprice.types import BaseModel, UNSET_SENTINEL
from pydantic import model_serializer
from typing import Optional
from typing_extensions import NotRequired, TypedDict


class DtoMatchedSubscriptionLineItemTypedDict(TypedDict):
    end_date: NotRequired[str]
    is_active_for_event: NotRequired[bool]
    price_id: NotRequired[str]
    start_date: NotRequired[str]
    sub_line_item_id: NotRequired[str]
    subscription_id: NotRequired[str]
    subscription_line_item: NotRequired[SubscriptionSubscriptionLineItemTypedDict]
    timestamp_within_range: NotRequired[bool]


class DtoMatchedSubscriptionLineItem(BaseModel):
    end_date: Optional[str] = None

    is_active_for_event: Optional[bool] = None

    price_id: Optional[str] = None

    start_date: Optional[str] = None

    sub_line_item_id: Optional[str] = None

    subscription_id: Optional[str] = None

    subscription_line_item: Optional[SubscriptionSubscriptionLineItem] = None

    timestamp_within_range: Optional[bool] = None

    @model_serializer(mode="wrap")
    def serialize_model(self, handler):
        optional_fields = set(
            [
                "end_date",
                "is_active_for_event",
                "price_id",
                "start_date",
                "sub_line_item_id",
                "subscription_id",
                "subscription_line_item",
                "timestamp_within_range",
            ]
        )
        serialized = handler(self)
        m = {}

        for n, f in type(self).model_fields.items():
            k = f.alias or n
            val = serialized.get(k, serialized.get(n))

            if val != UNSET_SENTINEL:
                if val is not None or k not in optional_fields:
                    m[k] = val

        return m



"""Code generated by Speakeasy (https://speakeasy.com). DO NOT EDIT."""

from __future__ import annotations
from typing import Literal


ScheduleType = Literal[
    "immediate",
    "end_of_period",
]



"""Code generated by Speakeasy (https://speakeasy.com). DO NOT EDIT."""

from __future__ import annotations
from .dto_userresponse import DtoUserResponse, DtoUserResponseTypedDict
from .paginationresponse import PaginationResponse, PaginationResponseTypedDict
from flexprice.types import BaseModel, UNSET_SENTINEL
from pydantic import model_serializer
from typing import List, Optional
from typing_extensions import NotRequired, TypedDict


class DtoListUsersResponseTypedDict(TypedDict):
    items: NotRequired[List[DtoUserResponseTypedDict]]
    pagination: NotRequired[PaginationResponseTypedDict]


class DtoListUsersResponse(BaseModel):
    items: Optional[List[DtoUserResponse]] = None

    pagination: Optional[PaginationResponse] = None

    @model_serializer(mode="wrap")
    def serialize_model(self, handler):
        optional_fields = set(["items", "pagination"])
        serialized = handler(self)
        m = {}

        for n, f in type(self).model_fields.items():
            k = f.alias or n
            val = serialized.get(k, serialized.get(n))

            if val != UNSET_SENTINEL:
                if val is not None or k not in optional_fields:
                    m[k] = val

        return m



"""Code generated by Speakeasy (https://speakeasy.com). DO NOT EDIT."""

from __future__ import annotations
from .dto_billingcycleinfo import DtoBillingCycleInfo, DtoBillingCycleInfoTypedDict
from .dto_invoicepreview import DtoInvoicePreview, DtoInvoicePreviewTypedDict
from .dto_plansummary import DtoPlanSummary, DtoPlanSummaryTypedDict
from .dto_prorationdetails import DtoProrationDetails, DtoProrationDetailsTypedDict
from .subscriptionchangetype import SubscriptionChangeType
from flexprice.types import BaseModel, UNSET_SENTINEL
from pydantic import model_serializer
from typing import Dict, List, Optional
from typing_extensions import NotRequired, TypedDict


class DtoSubscriptionChangePreviewResponseTypedDict(TypedDict):
    r"""Response showing the financial impact of a subscription plan change"""

    change_type: NotRequired[SubscriptionChangeType]
    current_plan: NotRequired[DtoPlanSummaryTypedDict]
    effective_date: NotRequired[str]
    r"""effective_date is when the change would take effect"""
    metadata: NotRequired[Dict[str, str]]
    r"""metadata from the request"""
    new_billing_cycle: NotRequired[DtoBillingCycleInfoTypedDict]
    next_invoice_preview: NotRequired[DtoInvoicePreviewTypedDict]
    proration_details: NotRequired[DtoProrationDetailsTypedDict]
    subscription_id: NotRequired[str]
    r"""subscription_id is the ID of the subscription being changed"""
    target_plan: NotRequired[DtoPlanSummaryTypedDict]
    warnings: NotRequired[List[str]]
    r"""warnings contains any warnings about the change"""


class DtoSubscriptionChangePreviewResponse(BaseModel):
    r"""Response showing the financial impact of a subscription plan change"""

    change_type: Optional[SubscriptionChangeType] = None

    current_plan: Optional[DtoPlanSummary] = None

    effective_date: Optional[str] = None
    r"""effective_date is when the change would take effect"""

    metadata: Optional[Dict[str, str]] = None
    r"""metadata from the request"""

    new_billing_cycle: Optional[DtoBillingCycleInfo] = None

    next_invoice_preview: Optional[DtoInvoicePreview] = None

    proration_details: Optional[DtoProrationDetails] = None

    subscription_id: Optional[str] = None
    r"""subscription_id is the ID of the subscription being changed"""

    target_plan: Optional[DtoPlanSummary] = None

    warnings: Optional[List[str]] = None
    r"""warnings contains any warnings about the change"""

    @model_serializer(mode="wrap")
    def serialize_model(self, handler):
        optional_fields = set(
            [
                "change_type",
                "current_plan",
                "effective_date",
                "metadata",
                "new_billing_cycle",
                "next_invoice_preview",
                "proration_details",
                "subscription_id",
                "target_plan",
                "warnings",
            ]
        )
        serialized = handler(self)
        m = {}

        for n, f in type(self).model_fields.items():
            k = f.alias or n
            val = serialized.get(k, serialized.get(n))

            if val != UNSET_SENTINEL:
                if val is not None or k not in optional_fields:
                    m[k] = val

        return m



"""Code generated by Speakeasy (https://speakeasy.com). DO NOT EDIT."""

from __future__ import annotations
from .billingmodel import BillingModel
from .billingtier import BillingTier
from .dto_createpricetier import DtoCreatePriceTier, DtoCreatePriceTierTypedDict
from .price_transformquantity import (
    PriceTransformQuantity,
    PriceTransformQuantityTypedDict,
)
from flexprice.types import BaseModel, UNSET_SENTINEL
from pydantic import model_serializer
from typing import Dict, List, Optional
from typing_extensions import NotRequired, TypedDict


class DtoUpdatePriceRequestTypedDict(TypedDict):
    amount: NotRequired[str]
    r"""Amount is the new price amount that overrides the original price (optional)"""
    billing_model: NotRequired[BillingModel]
    description: NotRequired[str]
    display_name: NotRequired[str]
    effective_from: NotRequired[str]
    group_id: NotRequired[str]
    r"""GroupID is the id of the group to update the price in.
    If not provided (nil), the group will not be changed
    If provided as empty string (\"\"), the group will be removed (price will be ungrouped)
    If provided as a group ID, the price will be assigned to that group (must exist and be published)
    """
    lookup_key: NotRequired[str]
    r"""All price fields that can be updated
    Non-critical fields (can be updated directly)
    """
    metadata: NotRequired[Dict[str, str]]
    price_unit_amount: NotRequired[str]
    r"""PriceUnitAmount is the price unit amount (for CUSTOM price unit type, FLAT_FEE/PACKAGE billing models)"""
    price_unit_tiers: NotRequired[List[DtoCreatePriceTierTypedDict]]
    r"""PriceUnitTiers are the price unit tiers (for CUSTOM price unit type, TIERED billing model)"""
    tier_mode: NotRequired[BillingTier]
    tiers: NotRequired[List[DtoCreatePriceTierTypedDict]]
    r"""Tiers determines the pricing tiers for this line item"""
    transform_quantity: NotRequired[PriceTransformQuantityTypedDict]


class DtoUpdatePriceRequest(BaseModel):
    amount: Optional[str] = None
    r"""Amount is the new price amount that overrides the original price (optional)"""

    billing_model: Optional[BillingModel] = None

    description: Optional[str] = None

    display_name: Optional[str] = None

    effective_from: Optional[str] = None

    group_id: Optional[str] = None
    r"""GroupID is the id of the group to update the price in.
    If not provided (nil), the group will not be changed
    If provided as empty string (\"\"), the group will be removed (price will be ungrouped)
    If provided as a group ID, the price will be assigned to that group (must exist and be published)
    """

    lookup_key: Optional[str] = None
    r"""All price fields that can be updated
    Non-critical fields (can be updated directly)
    """

    metadata: Optional[Dict[str, str]] = None

    price_unit_amount: Optional[str] = None
    r"""PriceUnitAmount is the price unit amount (for CUSTOM price unit type, FLAT_FEE/PACKAGE billing models)"""

    price_unit_tiers: Optional[List[DtoCreatePriceTier]] = None
    r"""PriceUnitTiers are the price unit tiers (for CUSTOM price unit type, TIERED billing model)"""

    tier_mode: Optional[BillingTier] = None

    tiers: Optional[List[DtoCreatePriceTier]] = None
    r"""Tiers determines the pricing tiers for this line item"""

    transform_quantity: Optional[PriceTransformQuantity] = None

    @model_serializer(mode="wrap")
    def serialize_model(self, handler):
        optional_fields = set(
            [
                "amount",
                "billing_model",
                "description",
                "display_name",
                "effective_from",
                "group_id",
                "lookup_key",
                "metadata",
                "price_unit_amount",
                "price_unit_tiers",
                "tier_mode",
                "tiers",
                "transform_quantity",
            ]
        )
        serialized = handler(self)
        m = {}

        for n, f in type(self).model_fields.items():
            k = f.alias or n
            val = serialized.get(k, serialized.get(n))

            if val != UNSET_SENTINEL:
                if val is not None or k not in optional_fields:
                    m[k] = val

        return m



"""Code generated by Speakeasy (https://speakeasy.com). DO NOT EDIT."""

from __future__ import annotations
from typing import Literal


CollectionMethod = Literal[
    "charge_automatically",
    "send_invoice",
]



"""Code generated by Speakeasy (https://speakeasy.com). DO NOT EDIT."""

from .basesdk import BaseSDK
from flexprice import models, utils
from flexprice._hooks import HookContext
from flexprice.types import OptionalNullable, UNSET
from flexprice.utils.unmarshal_json_response import unmarshal_json_response
from typing import Any, Dict, List, Mapping, Optional, Union


class Customers(BaseSDK):
    def update_customer(
        self,
        *,
        id: Optional[str] = None,
        external_customer_id: Optional[str] = None,
        address_city: Optional[str] = None,
        address_country: Optional[str] = None,
        address_line1: Optional[str] = None,
        address_line2: Optional[str] = None,
        address_postal_code: Optional[str] = None,
        address_state: Optional[str] = None,
        email: Optional[str] = None,
        external_id: Optional[str] = None,
        integration_entity_mapping: Optional[
            Union[
                List[models.DtoIntegrationEntityMapping],
                List[models.DtoIntegrationEntityMappingTypedDict],
            ]
        ] = None,
        metadata: Optional[Dict[str, str]] = None,
        name: Optional[str] = None,
        parent_customer_external_id: Optional[str] = None,
        parent_customer_id: Optional[str] = None,
        retries: OptionalNullable[utils.RetryConfig] = UNSET,
        server_url: Optional[str] = None,
        timeout_ms: Optional[int] = None,
        http_headers: Optional[Mapping[str, str]] = None,
    ) -> models.DtoCustomerResponse:
        r"""Update customer

        Use when updating customer details (e.g. name, email, or metadata). Identify by id or external_customer_id.

        :param id: Customer ID
        :param external_customer_id: Customer External ID
        :param address_city: address_city is the updated city name with maximum 100 characters
        :param address_country: address_country is the updated two-letter ISO 3166-1 alpha-2 country code
        :param address_line1: address_line1 is the updated primary address line with maximum 255 characters
        :param address_line2: address_line2 is the updated secondary address line with maximum 255 characters
        :param address_postal_code: address_postal_code is the updated postal code with maximum 20 characters
        :param address_state: address_state is the updated state, province, or region name with maximum 100 characters
        :param email: email is the updated email address and must be a valid email format if provided
        :param external_id: external_id is the updated external identifier for the customer
        :param integration_entity_mapping: integration_entity_mapping contains provider integration mappings for this customer
        :param metadata: metadata contains updated key-value pairs that will replace existing metadata
        :param name: name is the updated name or company name for the customer
        :param parent_customer_external_id: parent_customer_external_id is the external ID of the parent customer from your system
            Exactly one of parent_customer_id or parent_customer_external_id may be provided
            If you provide the external ID, the parent customer value will be ignored
        :param parent_customer_id: parent_customer_id is the internal FlexPrice ID of the parent customer
        :param retries: Override the default retry configuration for this method
        :param server_url: Override the default server URL for this method
        :param timeout_ms: Override the default request timeout configuration for this method in milliseconds
        :param http_headers: Additional headers to set or replace on requests.
        """
        base_url = None
        url_variables = None
        if timeout_ms is None:
            timeout_ms = self.sdk_configuration.timeout_ms

        if server_url is not None:
            base_url = server_url
        else:
            base_url = self._get_url(base_url, url_variables)

        request = models.UpdateCustomerRequest(
            id=id,
            external_customer_id=external_customer_id,
            body=models.DtoUpdateCustomerRequest(
                address_city=address_city,
                address_country=address_country,
                address_line1=address_line1,
                address_line2=address_line2,
                address_postal_code=address_postal_code,
                address_state=address_state,
                email=email,
                external_id=external_id,
                integration_entity_mapping=utils.get_pydantic_model(
                    integration_entity_mapping,
                    Optional[List[models.DtoIntegrationEntityMapping]],
                ),
                metadata=metadata,
                name=name,
                parent_customer_external_id=parent_customer_external_id,
                parent_customer_id=parent_customer_id,
            ),
        )

        req = self._build_request(
            method="PUT",
            path="/customers",
            base_url=base_url,
            url_variables=url_variables,
            request=request,
            request_body_required=True,
            request_has_path_params=False,
            request_has_query_params=True,
            user_agent_header="user-agent",
            accept_header_value="application/json",
            http_headers=http_headers,
            security=self.sdk_configuration.security,
            get_serialized_body=lambda: utils.serialize_request_body(
                request.body, False, False, "json", models.DtoUpdateCustomerRequest
            ),
            allow_empty_value=None,
            timeout_ms=timeout_ms,
        )

        if retries == UNSET:
            if self.sdk_configuration.retry_config is not UNSET:
                retries = self.sdk_configuration.retry_config

        retry_config = None
        if isinstance(retries, utils.RetryConfig):
            retry_config = (retries, ["429", "500", "502", "503", "504"])

        http_res = self.do_request(
            hook_ctx=HookContext(
                config=self.sdk_configuration,
                base_url=base_url or "",
                operation_id="updateCustomer",
                oauth2_scopes=None,
                security_source=self.sdk_configuration.security,
            ),
            request=req,
            error_status_codes=["400", "4XX", "500", "5XX"],
            retry_config=retry_config,
        )

        response_data: Any = None
        if utils.match_response(http_res, "200", "application/json"):
            return unmarshal_json_response(models.DtoCustomerResponse, http_res)
        if utils.match_response(http_res, "400", "application/json"):
            response_data = unmarshal_json_response(
                models.errors.ErrorsErrorResponseData, http_res
            )
            raise models.errors.ErrorsErrorResponse(response_data, http_res)
        if utils.match_response(http_res, "500", "application/json"):
            response_data = unmarshal_json_response(
                models.errors.ErrorsErrorResponseData, http_res
            )
            raise models.errors.ErrorsErrorResponse(response_data, http_res)
        if utils.match_response(http_res, "4XX", "*"):
            http_res_text = utils.stream_to_text(http_res)
            raise models.errors.FlexpriceDefaultError(
                "API error occurred", http_res, http_res_text
            )
        if utils.match_response(http_res, "5XX", "*"):
            http_res_text = utils.stream_to_text(http_res)
            raise models.errors.FlexpriceDefaultError(
                "API error occurred", http_res, http_res_text
            )

        raise models.errors.FlexpriceDefaultError(
            "Unexpected response received", http_res
        )

    async def update_customer_async(
        self,
        *,
        id: Optional[str] = None,
        external_customer_id: Optional[str] = None,
        address_city: Optional[str] = None,
        address_country: Optional[str] = None,
        address_line1: Optional[str] = None,
        address_line2: Optional[str] = None,
        address_postal_code: Optional[str] = None,
        address_state: Optional[str] = None,
        email: Optional[str] = None,
        external_id: Optional[str] = None,
        integration_entity_mapping: Optional[
            Union[
                List[models.DtoIntegrationEntityMapping],
                List[models.DtoIntegrationEntityMappingTypedDict],
            ]
        ] = None,
        metadata: Optional[Dict[str, str]] = None,
        name: Optional[str] = None,
        parent_customer_external_id: Optional[str] = None,
        parent_customer_id: Optional[str] = None,
        retries: OptionalNullable[utils.RetryConfig] = UNSET,
        server_url: Optional[str] = None,
        timeout_ms: Optional[int] = None,
        http_headers: Optional[Mapping[str, str]] = None,
    ) -> models.DtoCustomerResponse:
        r"""Update customer

        Use when updating customer details (e.g. name, email, or metadata). Identify by id or external_customer_id.

        :param id: Customer ID
        :param external_customer_id: Customer External ID
        :param address_city: address_city is the updated city name with maximum 100 characters
        :param address_country: address_country is the updated two-letter ISO 3166-1 alpha-2 country code
        :param address_line1: address_line1 is the updated primary address line with maximum 255 characters
        :param address_line2: address_line2 is the updated secondary address line with maximum 255 characters
        :param address_postal_code: address_postal_code is the updated postal code with maximum 20 characters
        :param address_state: address_state is the updated state, province, or region name with maximum 100 characters
        :param email: email is the updated email address and must be a valid email format if provided
        :param external_id: external_id is the updated external identifier for the customer
        :param integration_entity_mapping: integration_entity_mapping contains provider integration mappings for this customer
        :param metadata: metadata contains updated key-value pairs that will replace existing metadata
        :param name: name is the updated name or company name for the customer
        :param parent_customer_external_id: parent_customer_external_id is the external ID of the parent customer from your system
            Exactly one of parent_customer_id or parent_customer_external_id may be provided
            If you provide the external ID, the parent customer value will be ignored
        :param parent_customer_id: parent_customer_id is the internal FlexPrice ID of the parent customer
        :param retries: Override the default retry configuration for this method
        :param server_url: Override the default server URL for this method
        :param timeout_ms: Override the default request timeout configuration for this method in milliseconds
        :param http_headers: Additional headers to set or replace on requests.
        """
        base_url = None
        url_variables = None
        if timeout_ms is None:
            timeout_ms = self.sdk_configuration.timeout_ms

        if server_url is not None:
            base_url = server_url
        else:
            base_url = self._get_url(base_url, url_variables)

        request = models.UpdateCustomerRequest(
            id=id,
            external_customer_id=external_customer_id,
            body=models.DtoUpdateCustomerRequest(
                address_city=address_city,
                address_country=address_country,
                address_line1=address_line1,
                address_line2=address_line2,
                address_postal_code=address_postal_code,
                address_state=address_state,
                email=email,
                external_id=external_id,
                integration_entity_mapping=utils.get_pydantic_model(
                    integration_entity_mapping,
                    Optional[List[models.DtoIntegrationEntityMapping]],
                ),
                metadata=metadata,
                name=name,
                parent_customer_external_id=parent_customer_external_id,
                parent_customer_id=parent_customer_id,
            ),
        )

        req = self._build_request_async(
            method="PUT",
            path="/customers",
            base_url=base_url,
            url_variables=url_variables,
            request=request,
            request_body_required=True,
            request_has_path_params=False,
            request_has_query_params=True,
            user_agent_header="user-agent",
            accept_header_value="application/json",
            http_headers=http_headers,
            security=self.sdk_configuration.security,
            get_serialized_body=lambda: utils.serialize_request_body(
                request.body, False, False, "json", models.DtoUpdateCustomerRequest
            ),
            allow_empty_value=None,
            timeout_ms=timeout_ms,
        )

        if retries == UNSET:
            if self.sdk_configuration.retry_config is not UNSET:
                retries = self.sdk_configuration.retry_config

        retry_config = None
        if isinstance(retries, utils.RetryConfig):
            retry_config = (retries, ["429", "500", "502", "503", "504"])

        http_res = await self.do_request_async(
            hook_ctx=HookContext(
                config=self.sdk_configuration,
                base_url=base_url or "",
                operation_id="updateCustomer",
                oauth2_scopes=None,
                security_source=self.sdk_configuration.security,
            ),
            request=req,
            error_status_codes=["400", "4XX", "500", "5XX"],
            retry_config=retry_config,
        )

        response_data: Any = None
        if utils.match_response(http_res, "200", "application/json"):
            return unmarshal_json_response(models.DtoCustomerResponse, http_res)
        if utils.match_response(http_res, "400", "application/json"):
            response_data = unmarshal_json_response(
                models.errors.ErrorsErrorResponseData, http_res
            )
            raise models.errors.ErrorsErrorResponse(response_data, http_res)
        if utils.match_response(http_res, "500", "application/json"):
            response_data = unmarshal_json_response(
                models.errors.ErrorsErrorResponseData, http_res
            )
            raise models.errors.ErrorsErrorResponse(response_data, http_res)
        if utils.match_response(http_res, "4XX", "*"):
            http_res_text = await utils.stream_to_text_async(http_res)
            raise models.errors.FlexpriceDefaultError(
                "API error occurred", http_res, http_res_text
            )
        if utils.match_response(http_res, "5XX", "*"):
            http_res_text = await utils.stream_to_text_async(http_res)
            raise models.errors.FlexpriceDefaultError(
                "API error occurred", http_res, http_res_text
            )

        raise models.errors.FlexpriceDefaultError(
            "Unexpected response received", http_res
        )

    def create_customer(
        self,
        *,
        external_id: str,
        address_city: Optional[str] = None,
        address_country: Optional[str] = None,
        address_line1: Optional[str] = None,
        address_line2: Optional[str] = None,
        address_postal_code: Optional[str] = None,
        address_state: Optional[str] = None,
        email: Optional[str] = None,
        integration_entity_mapping: Optional[
            Union[
                List[models.DtoIntegrationEntityMapping],
                List[models.DtoIntegrationEntityMappingTypedDict],
            ]
        ] = None,
        metadata: Optional[Dict[str, str]] = None,
        name: Optional[str] = None,
        parent_customer_external_id: Optional[str] = None,
        parent_customer_id: Optional[str] = None,
        skip_onboarding_workflow: Optional[bool] = None,
        tax_rate_overrides: Optional[
            Union[
                List[models.DtoTaxRateOverride],
                List[models.DtoTaxRateOverrideTypedDict],
            ]
        ] = None,
        retries: OptionalNullable[utils.RetryConfig] = UNSET,
        server_url: Optional[str] = None,
        timeout_ms: Optional[int] = None,
        http_headers: Optional[Mapping[str, str]] = None,
    ) -> models.DtoCustomerResponse:
        r"""Create customer

        Use when onboarding a new billing customer (e.g. sign-up or CRM sync). Ideal for linking via external_customer_id to your app's user id.

        :param external_id: external_id is the unique identifier from your system to reference this customer (required)
        :param address_city: address_city is the city name with maximum 100 characters
        :param address_country: address_country is the two-letter ISO 3166-1 alpha-2 country code
        :param address_line1: address_line1 is the primary address line with maximum 255 characters
        :param address_line2: address_line2 is the secondary address line with maximum 255 characters
        :param address_postal_code: address_postal_code is the ZIP code or postal code with maximum 20 characters
        :param address_state: address_state is the state, province, or region name with maximum 100 characters
        :param email: email is the customer's email address and must be a valid email format if provided
        :param integration_entity_mapping: integration_entity_mapping contains provider integration mappings for this customer
        :param metadata: metadata contains additional key-value pairs for storing extra information
        :param name: name is the full name or company name of the customer
        :param parent_customer_external_id: parent_customer_external_id is the external ID of the parent customer from your system
            Exactly one of parent_customer_id or parent_customer_external_id may be provided
        :param parent_customer_id: parent_customer_id is the internal FlexPrice ID of the parent customer
        :param skip_onboarding_workflow: skip_onboarding_workflow when true, prevents the customer onboarding workflow from being triggered
            This is used internally when a customer is created via a workflow to prevent infinite loops
            Default: false
        :param tax_rate_overrides: tax_rate_overrides contains tax rate configurations to be linked to this customer
        :param retries: Override the default retry configuration for this method
        :param server_url: Override the default server URL for this method
        :param timeout_ms: Override the default request timeout configuration for this method in milliseconds
        :param http_headers: Additional headers to set or replace on requests.
        """
        base_url = None
        url_variables = None
        if timeout_ms is None:
            timeout_ms = self.sdk_configuration.timeout_ms

        if server_url is not None:
            base_url = server_url
        else:
            base_url = self._get_url(base_url, url_variables)

        request = models.DtoCreateCustomerRequest(
            address_city=address_city,
            address_country=address_country,
            address_line1=address_line1,
            address_line2=address_line2,
            address_postal_code=address_postal_code,
            address_state=address_state,
            email=email,
            external_id=external_id,
            integration_entity_mapping=utils.get_pydantic_model(
                integration_entity_mapping,
                Optional[List[models.DtoIntegrationEntityMapping]],
            ),
            metadata=metadata,
            name=name,
            parent_customer_external_id=parent_customer_external_id,
            parent_customer_id=parent_customer_id,
            skip_onboarding_workflow=skip_onboarding_workflow,
            tax_rate_overrides=utils.get_pydantic_model(
                tax_rate_overrides, Optional[List[models.DtoTaxRateOverride]]
            ),
        )

        req = self._build_request(
            method="POST",
            path="/customers",
            base_url=base_url,
            url_variables=url_variables,
            request=request,
            request_body_required=True,
            request_has_path_params=False,
            request_has_query_params=True,
            user_agent_header="user-agent",
            accept_header_value="application/json",
            http_headers=http_headers,
            security=self.sdk_configuration.security,
            get_serialized_body=lambda: utils.serialize_request_body(
                request, False, False, "json", models.DtoCreateCustomerRequest
            ),
            allow_empty_value=None,
            timeout_ms=timeout_ms,
        )

        if retries == UNSET:
            if self.sdk_configuration.retry_config is not UNSET:
                retries = self.sdk_configuration.retry_config

        retry_config = None
        if isinstance(retries, utils.RetryConfig):
            retry_config = (retries, ["429", "500", "502", "503", "504"])

        http_res = self.do_request(
            hook_ctx=HookContext(
                config=self.sdk_configuration,
                base_url=base_url or "",
                operation_id="createCustomer",
                oauth2_scopes=None,
                security_source=self.sdk_configuration.security,
            ),
            request=req,
            error_status_codes=["400", "4XX", "500", "5XX"],
            retry_config=retry_config,
        )

        response_data: Any = None
        if utils.match_response(http_res, "201", "application/json"):
            return unmarshal_json_response(models.DtoCustomerResponse, http_res)
        if utils.match_response(http_res, "400", "application/json"):
            response_data = unmarshal_json_response(
                models.errors.ErrorsErrorResponseData, http_res
            )
            raise models.errors.ErrorsErrorResponse(response_data, http_res)
        if utils.match_response(http_res, "500", "application/json"):
            response_data = unmarshal_json_response(
                models.errors.ErrorsErrorResponseData, http_res
            )
            raise models.errors.ErrorsErrorResponse(response_data, http_res)
        if utils.match_response(http_res, "4XX", "*"):
            http_res_text = utils.stream_to_text(http_res)
            raise models.errors.FlexpriceDefaultError(
                "API error occurred", http_res, http_res_text
            )
        if utils.match_response(http_res, "5XX", "*"):
            http_res_text = utils.stream_to_text(http_res)
            raise models.errors.FlexpriceDefaultError(
                "API error occurred", http_res, http_res_text
            )

        raise models.errors.FlexpriceDefaultError(
            "Unexpected response received", http_res
        )

    async def create_customer_async(
        self,
        *,
        external_id: str,
        address_city: Optional[str] = None,
        address_country: Optional[str] = None,
        address_line1: Optional[str] = None,
        address_line2: Optional[str] = None,
        address_postal_code: Optional[str] = None,
        address_state: Optional[str] = None,
        email: Optional[str] = None,
        integration_entity_mapping: Optional[
            Union[
                List[models.DtoIntegrationEntityMapping],
                List[models.DtoIntegrationEntityMappingTypedDict],
            ]
        ] = None,
        metadata: Optional[Dict[str, str]] = None,
        name: Optional[str] = None,
        parent_customer_external_id: Optional[str] = None,
        parent_customer_id: Optional[str] = None,
        skip_onboarding_workflow: Optional[bool] = None,
        tax_rate_overrides: Optional[
            Union[
                List[models.DtoTaxRateOverride],
                List[models.DtoTaxRateOverrideTypedDict],
            ]
        ] = None,
        retries: OptionalNullable[utils.RetryConfig] = UNSET,
        server_url: Optional[str] = None,
        timeout_ms: Optional[int] = None,
        http_headers: Optional[Mapping[str, str]] = None,
    ) -> models.DtoCustomerResponse:
        r"""Create customer

        Use when onboarding a new billing customer (e.g. sign-up or CRM sync). Ideal for linking via external_customer_id to your app's user id.

        :param external_id: external_id is the unique identifier from your system to reference this customer (required)
        :param address_city: address_city is the city name with maximum 100 characters
        :param address_country: address_country is the two-letter ISO 3166-1 alpha-2 country code
        :param address_line1: address_line1 is the primary address line with maximum 255 characters
        :param address_line2: address_line2 is the secondary address line with maximum 255 characters
        :param address_postal_code: address_postal_code is the ZIP code or postal code with maximum 20 characters
        :param address_state: address_state is the state, province, or region name with maximum 100 characters
        :param email: email is the customer's email address and must be a valid email format if provided
        :param integration_entity_mapping: integration_entity_mapping contains provider integration mappings for this customer
        :param metadata: metadata contains additional key-value pairs for storing extra information
        :param name: name is the full name or company name of the customer
        :param parent_customer_external_id: parent_customer_external_id is the external ID of the parent customer from your system
            Exactly one of parent_customer_id or parent_customer_external_id may be provided
        :param parent_customer_id: parent_customer_id is the internal FlexPrice ID of the parent customer
        :param skip_onboarding_workflow: skip_onboarding_workflow when true, prevents the customer onboarding workflow from being triggered
            This is used internally when a customer is created via a workflow to prevent infinite loops
            Default: false
        :param tax_rate_overrides: tax_rate_overrides contains tax rate configurations to be linked to this customer
        :param retries: Override the default retry configuration for this method
        :param server_url: Override the default server URL for this method
        :param timeout_ms: Override the default request timeout configuration for this method in milliseconds
        :param http_headers: Additional headers to set or replace on requests.
        """
        base_url = None
        url_variables = None
        if timeout_ms is None:
            timeout_ms = self.sdk_configuration.timeout_ms

        if server_url is not None:
            base_url = server_url
        else:
            base_url = self._get_url(base_url, url_variables)

        request = models.DtoCreateCustomerRequest(
            address_city=address_city,
            address_country=address_country,
            address_line1=address_line1,
            address_line2=address_line2,
            address_postal_code=address_postal_code,
            address_state=address_state,
            email=email,
            external_id=external_id,
            integration_entity_mapping=utils.get_pydantic_model(
                integration_entity_mapping,
                Optional[List[models.DtoIntegrationEntityMapping]],
            ),
            metadata=metadata,
            name=name,
            parent_customer_external_id=parent_customer_external_id,
            parent_customer_id=parent_customer_id,
            skip_onboarding_workflow=skip_onboarding_workflow,
            tax_rate_overrides=utils.get_pydantic_model(
                tax_rate_overrides, Optional[List[models.DtoTaxRateOverride]]
            ),
        )

        req = self._build_request_async(
            method="POST",
            path="/customers",
            base_url=base_url,
            url_variables=url_variables,
            request=request,
            request_body_required=True,
            request_has_path_params=False,
            request_has_query_params=True,
            user_agent_header="user-agent",
            accept_header_value="application/json",
            http_headers=http_headers,
            security=self.sdk_configuration.security,
            get_serialized_body=lambda: utils.serialize_request_body(
                request, False, False, "json", models.DtoCreateCustomerRequest
            ),
            allow_empty_value=None,
            timeout_ms=timeout_ms,
        )

        if retries == UNSET:
            if self.sdk_configuration.retry_config is not UNSET:
                retries = self.sdk_configuration.retry_config

        retry_config = None
        if isinstance(retries, utils.RetryConfig):
            retry_config = (retries, ["429", "500", "502", "503", "504"])

        http_res = await self.do_request_async(
            hook_ctx=HookContext(
                config=self.sdk_configuration,
                base_url=base_url or "",
                operation_id="createCustomer",
                oauth2_scopes=None,
                security_source=self.sdk_configuration.security,
            ),
            request=req,
            error_status_codes=["400", "4XX", "500", "5XX"],
            retry_config=retry_config,
        )

        response_data: Any = None
        if utils.match_response(http_res, "201", "application/json"):
            return unmarshal_json_response(models.DtoCustomerResponse, http_res)
        if utils.match_response(http_res, "400", "application/json"):
            response_data = unmarshal_json_response(
                models.errors.ErrorsErrorResponseData, http_res
            )
            raise models.errors.ErrorsErrorResponse(response_data, http_res)
        if utils.match_response(http_res, "500", "application/json"):
            response_data = unmarshal_json_response(
                models.errors.ErrorsErrorResponseData, http_res
            )
            raise models.errors.ErrorsErrorResponse(response_data, http_res)
        if utils.match_response(http_res, "4XX", "*"):
            http_res_text = await utils.stream_to_text_async(http_res)
            raise models.errors.FlexpriceDefaultError(
                "API error occurred", http_res, http_res_text
            )
        if utils.match_response(http_res, "5XX", "*"):
            http_res_text = await utils.stream_to_text_async(http_res)
            raise models.errors.FlexpriceDefaultError(
                "API error occurred", http_res, http_res_text
            )

        raise models.errors.FlexpriceDefaultError(
            "Unexpected response received", http_res
        )

    def get_customer_by_external_id(
        self,
        *,
        external_id: str,
        retries: OptionalNullable[utils.RetryConfig] = UNSET,
        server_url: Optional[str] = None,
        timeout_ms: Optional[int] = None,
        http_headers: Optional[Mapping[str, str]] = None,
    ) -> models.DtoCustomerResponse:
        r"""Get customer by external ID

        Use when resolving a customer by your app's id (e.g. from your user table). Ideal for integrations that key by external id.

        :param external_id: Customer External ID
        :param retries: Override the default retry configuration for this method
        :param server_url: Override the default server URL for this method
        :param timeout_ms: Override the default request timeout configuration for this method in milliseconds
        :param http_headers: Additional headers to set or replace on requests.
        """
        base_url = None
        url_variables = None
        if timeout_ms is None:
            timeout_ms = self.sdk_configuration.timeout_ms

        if server_url is not None:
            base_url = server_url
        else:
            base_url = self._get_url(base_url, url_variables)

        request = models.GetCustomerByExternalIDRequest(
            external_id=external_id,
        )

        req = self._build_request(
            method="GET",
            path="/customers/external/{external_id}",
            base_url=base_url,
            url_variables=url_variables,
            request=request,
            request_body_required=False,
            request_has_path_params=True,
            request_has_query_params=True,
            user_agent_header="user-agent",
            accept_header_value="application/json",
            http_headers=http_headers,
            security=self.sdk_configuration.security,
            allow_empty_value=None,
            timeout_ms=timeout_ms,
        )

        if retries == UNSET:
            if self.sdk_configuration.retry_config is not UNSET:
                retries = self.sdk_configuration.retry_config

        retry_config = None
        if isinstance(retries, utils.RetryConfig):
            retry_config = (retries, ["429", "500", "502", "503", "504"])

        http_res = self.do_request(
            hook_ctx=HookContext(
                config=self.sdk_configuration,
                base_url=base_url or "",
                operation_id="getCustomerByExternalId",
                oauth2_scopes=None,
                security_source=self.sdk_configuration.security,
            ),
            request=req,
            error_status_codes=["400", "404", "4XX", "500", "5XX"],
            retry_config=retry_config,
        )

        response_data: Any = None
        if utils.match_response(http_res, "200", "application/json"):
            return unmarshal_json_response(models.DtoCustomerResponse, http_res)
        if utils.match_response(http_res, ["400", "404"], "application/json"):
            response_data = unmarshal_json_response(
                models.errors.ErrorsErrorResponseData, http_res
            )
            raise models.errors.ErrorsErrorResponse(response_data, http_res)
        if utils.match_response(http_res, "500", "application/json"):
            response_data = unmarshal_json_response(
                models.errors.ErrorsErrorResponseData, http_res
            )
            raise models.errors.ErrorsErrorResponse(response_data, http_res)
        if utils.match_response(http_res, "4XX", "*"):
            http_res_text = utils.stream_to_text(http_res)
            raise models.errors.FlexpriceDefaultError(
                "API error occurred", http_res, http_res_text
            )
        if utils.match_response(http_res, "5XX", "*"):
            http_res_text = utils.stream_to_text(http_res)
            raise models.errors.FlexpriceDefaultError(
                "API error occurred", http_res, http_res_text
            )

        raise models.errors.FlexpriceDefaultError(
            "Unexpected response received", http_res
        )

    async def get_customer_by_external_id_async(
        self,
        *,
        external_id: str,
        retries: OptionalNullable[utils.RetryConfig] = UNSET,
        server_url: Optional[str] = None,
        timeout_ms: Optional[int] = None,
        http_headers: Optional[Mapping[str, str]] = None,
    ) -> models.DtoCustomerResponse:
        r"""Get customer by external ID

        Use when resolving a customer by your app's id (e.g. from your user table). Ideal for integrations that key by external id.

        :param external_id: Customer External ID
        :param retries: Override the default retry configuration for this method
        :param server_url: Override the default server URL for this method
        :param timeout_ms: Override the default request timeout configuration for this method in milliseconds
        :param http_headers: Additional headers to set or replace on requests.
        """
        base_url = None
        url_variables = None
        if timeout_ms is None:
            timeout_ms = self.sdk_configuration.timeout_ms

        if server_url is not None:
            base_url = server_url
        else:
            base_url = self._get_url(base_url, url_variables)

        request = models.GetCustomerByExternalIDRequest(
            external_id=external_id,
        )

        req = self._build_request_async(
            method="GET",
            path="/customers/external/{external_id}",
            base_url=base_url,
            url_variables=url_variables,
            request=request,
            request_body_required=False,
            request_has_path_params=True,
            request_has_query_params=True,
            user_agent_header="user-agent",
            accept_header_value="application/json",
            http_headers=http_headers,
            security=self.sdk_configuration.security,
            allow_empty_value=None,
            timeout_ms=timeout_ms,
        )

        if retries == UNSET:
            if self.sdk_configuration.retry_config is not UNSET:
                retries = self.sdk_configuration.retry_config

        retry_config = None
        if isinstance(retries, utils.RetryConfig):
            retry_config = (retries, ["429", "500", "502", "503", "504"])

        http_res = await self.do_request_async(
            hook_ctx=HookContext(
                config=self.sdk_configuration,
                base_url=base_url or "",
                operation_id="getCustomerByExternalId",
                oauth2_scopes=None,
                security_source=self.sdk_configuration.security,
            ),
            request=req,
            error_status_codes=["400", "404", "4XX", "500", "5XX"],
            retry_config=retry_config,
        )

        response_data: Any = None
        if utils.match_response(http_res, "200", "application/json"):
            return unmarshal_json_response(models.DtoCustomerResponse, http_res)
        if utils.match_response(http_res, ["400", "404"], "application/json"):
            response_data = unmarshal_json_response(
                models.errors.ErrorsErrorResponseData, http_res
            )
            raise models.errors.ErrorsErrorResponse(response_data, http_res)
        if utils.match_response(http_res, "500", "application/json"):
            response_data = unmarshal_json_response(
                models.errors.ErrorsErrorResponseData, http_res
            )
            raise models.errors.ErrorsErrorResponse(response_data, http_res)
        if utils.match_response(http_res, "4XX", "*"):
            http_res_text = await utils.stream_to_text_async(http_res)
            raise models.errors.FlexpriceDefaultError(
                "API error occurred", http_res, http_res_text
            )
        if utils.match_response(http_res, "5XX", "*"):
            http_res_text = await utils.stream_to_text_async(http_res)
            raise models.errors.FlexpriceDefaultError(
                "API error occurred", http_res, http_res_text
            )

        raise models.errors.FlexpriceDefaultError(
            "Unexpected response received", http_res
        )

    def query_customer(
        self,
        *,
        customer_ids: Optional[List[str]] = None,
        email: Optional[str] = None,
        end_time: Optional[str] = None,
        expand: Optional[str] = None,
        external_id: Optional[str] = None,
        external_ids: Optional[List[str]] = None,
        filters: Optional[
            Union[List[models.FilterCondition], List[models.FilterConditionTypedDict]]
        ] = None,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
        order: Optional[models.CustomerFilterOrder] = None,
        parent_customer_ids: Optional[List[str]] = None,
        sort: Optional[
            Union[List[models.SortCondition], List[models.SortConditionTypedDict]]
        ] = None,
        start_time: Optional[str] = None,
        status: Optional[models.Status] = None,
        retries: OptionalNullable[utils.RetryConfig] = UNSET,
        server_url: Optional[str] = None,
        timeout_ms: Optional[int] = None,
        http_headers: Optional[Mapping[str, str]] = None,
    ) -> models.DtoListCustomersResponse:
        r"""Query customers

        Use when listing or searching customers (e.g. admin CRM or reporting). Returns a paginated list; supports filtering and sorting.

        :param customer_ids:
        :param email:
        :param end_time:
        :param expand:
        :param external_id:
        :param external_ids:
        :param filters:
        :param limit:
        :param offset:
        :param order:
        :param parent_customer_ids:
        :param sort:
        :param start_time:
        :param status:
        :param retries: Override the default retry configuration for this method
        :param server_url: Override the default server URL for this method
        :param timeout_ms: Override the default request timeout configuration for this method in milliseconds
        :param http_headers: Additional headers to set or replace on requests.
        """
        base_url = None
        url_variables = None
        if timeout_ms is None:
            timeout_ms = self.sdk_configuration.timeout_ms

        if server_url is not None:
            base_url = server_url
        else:
            base_url = self._get_url(base_url, url_variables)

        request = models.CustomerFilter(
            customer_ids=customer_ids,
            email=email,
            end_time=end_time,
            expand=expand,
            external_id=external_id,
            external_ids=external_ids,
            filters=utils.get_pydantic_model(
                filters, Optional[List[models.FilterCondition]]
            ),
            limit=limit,
            offset=offset,
            order=order,
            parent_customer_ids=parent_customer_ids,
            sort=utils.get_pydantic_model(sort, Optional[List[models.SortCondition]]),
            start_time=start_time,
            status=status,
        )

        req = self._build_request(
            method="POST",
            path="/customers/search",
            base_url=base_url,
            url_variables=url_variables,
            request=request,
            request_body_required=True,
            request_has_path_params=False,
            request_has_query_params=True,
            user_agent_header="user-agent",
            accept_header_value="application/json",
            http_headers=http_headers,
            security=self.sdk_configuration.security,
            get_serialized_body=lambda: utils.serialize_request_body(
                request, False, False, "json", models.CustomerFilter
            ),
            allow_empty_value=None,
            timeout_ms=timeout_ms,
        )

        if retries == UNSET:
            if self.sdk_configuration.retry_config is not UNSET:
                retries = self.sdk_configuration.retry_config

        retry_config = None
        if isinstance(retries, utils.RetryConfig):
            retry_config = (retries, ["429", "500", "502", "503", "504"])

        http_res = self.do_request(
            hook_ctx=HookContext(
                config=self.sdk_configuration,
                base_url=base_url or "",
                operation_id="queryCustomer",
                oauth2_scopes=None,
                security_source=self.sdk_configuration.security,
            ),
            request=req,
            error_status_codes=["400", "4XX", "500", "5XX"],
            retry_config=retry_config,
        )

        response_data: Any = None
        if utils.match_response(http_res, "200", "application/json"):
            return unmarshal_json_response(models.DtoListCustomersResponse, http_res)
        if utils.match_response(http_res, "400", "application/json"):
            response_data = unmarshal_json_response(
                models.errors.ErrorsErrorResponseData, http_res
            )
            raise models.errors.ErrorsErrorResponse(response_data, http_res)
        if utils.match_response(http_res, "500", "application/json"):
            response_data = unmarshal_json_response(
                models.errors.ErrorsErrorResponseData, http_res
            )
            raise models.errors.ErrorsErrorResponse(response_data, http_res)
        if utils.match_response(http_res, "4XX", "*"):
            http_res_text = utils.stream_to_text(http_res)
            raise models.errors.FlexpriceDefaultError(
                "API error occurred", http_res, http_res_text
            )
        if utils.match_response(http_res, "5XX", "*"):
            http_res_text = utils.stream_to_text(http_res)
            raise models.errors.FlexpriceDefaultError(
                "API error occurred", http_res, http_res_text
            )

        raise models.errors.FlexpriceDefaultError(
            "Unexpected response received", http_res
        )

    async def query_customer_async(
        self,
        *,
        customer_ids: Optional[List[str]] = None,
        email: Optional[str] = None,
        end_time: Optional[str] = None,
        expand: Optional[str] = None,
        external_id: Optional[str] = None,
        external_ids: Optional[List[str]] = None,
        filters: Optional[
            Union[List[models.FilterCondition], List[models.FilterConditionTypedDict]]
        ] = None,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
        order: Optional[models.CustomerFilterOrder] = None,
        parent_customer_ids: Optional[List[str]] = None,
        sort: Optional[
            Union[List[models.SortCondition], List[models.SortConditionTypedDict]]
        ] = None,
        start_time: Optional[str] = None,
        status: Optional[models.Status] = None,
        retries: OptionalNullable[utils.RetryConfig] = UNSET,
        server_url: Optional[str] = None,
        timeout_ms: Optional[int] = None,
        http_headers: Optional[Mapping[str, str]] = None,
    ) -> models.DtoListCustomersResponse:
        r"""Query customers

        Use when listing or searching customers (e.g. admin CRM or reporting). Returns a paginated list; supports filtering and sorting.

        :param customer_ids:
        :param email:
        :param end_time:
        :param expand:
        :param external_id:
        :param external_ids:
        :param filters:
        :param limit:
        :param offset:
        :param order:
        :param parent_customer_ids:
        :param sort:
        :param start_time:
        :param status:
        :param retries: Override the default retry configuration for this method
        :param server_url: Override the default server URL for this method
        :param timeout_ms: Override the default request timeout configuration for this method in milliseconds
        :param http_headers: Additional headers to set or replace on requests.
        """
        base_url = None
        url_variables = None
        if timeout_ms is None:
            timeout_ms = self.sdk_configuration.timeout_ms

        if server_url is not None:
            base_url = server_url
        else:
            base_url = self._get_url(base_url, url_variables)

        request = models.CustomerFilter(
            customer_ids=customer_ids,
            email=email,
            end_time=end_time,
            expand=expand,
            external_id=external_id,
            external_ids=external_ids,
            filters=utils.get_pydantic_model(
                filters, Optional[List[models.FilterCondition]]
            ),
            limit=limit,
            offset=offset,
            order=order,
            parent_customer_ids=parent_customer_ids,
            sort=utils.get_pydantic_model(sort, Optional[List[models.SortCondition]]),
            start_time=start_time,
            status=status,
        )

        req = self._build_request_async(
            method="POST",
            path="/customers/search",
            base_url=base_url,
            url_variables=url_variables,
            request=request,
            request_body_required=True,
            request_has_path_params=False,
            request_has_query_params=True,
            user_agent_header="user-agent",
            accept_header_value="application/json",
            http_headers=http_headers,
            security=self.sdk_configuration.security,
            get_serialized_body=lambda: utils.serialize_request_body(
                request, False, False, "json", models.CustomerFilter
            ),
            allow_empty_value=None,
            timeout_ms=timeout_ms,
        )

        if retries == UNSET:
            if self.sdk_configuration.retry_config is not UNSET:
                retries = self.sdk_configuration.retry_config

        retry_config = None
        if isinstance(retries, utils.RetryConfig):
            retry_config = (retries, ["429", "500", "502", "503", "504"])

        http_res = await self.do_request_async(
            hook_ctx=HookContext(
                config=self.sdk_configuration,
                base_url=base_url or "",
                operation_id="queryCustomer",
                oauth2_scopes=None,
                security_source=self.sdk_configuration.security,
            ),
            request=req,
            error_status_codes=["400", "4XX", "500", "5XX"],
            retry_config=retry_config,
        )

        response_data: Any = None
        if utils.match_response(http_res, "200", "application/json"):
            return unmarshal_json_response(models.DtoListCustomersResponse, http_res)
        if utils.match_response(http_res, "400", "application/json"):
            response_data = unmarshal_json_response(
                models.errors.ErrorsErrorResponseData, http_res
            )
            raise models.errors.ErrorsErrorResponse(response_data, http_res)
        if utils.match_response(http_res, "500", "application/json"):
            response_data = unmarshal_json_response(
                models.errors.ErrorsErrorResponseData, http_res
            )
            raise models.errors.ErrorsErrorResponse(response_data, http_res)
        if utils.match_response(http_res, "4XX", "*"):
            http_res_text = await utils.stream_to_text_async(http_res)
            raise models.errors.FlexpriceDefaultError(
                "API error occurred", http_res, http_res_text
            )
        if utils.match_response(http_res, "5XX", "*"):
            http_res_text = await utils.stream_to_text_async(http_res)
            raise models.errors.FlexpriceDefaultError(
                "API error occurred", http_res, http_res_text
            )

        raise models.errors.FlexpriceDefaultError(
            "Unexpected response received", http_res
        )

    def get_customer_usage_summary(
        self,
        *,
        customer_id: Optional[str] = None,
        customer_lookup_key: Optional[str] = None,
        feature_ids: Optional[List[str]] = None,
        feature_lookup_keys: Optional[List[str]] = None,
        subscription_ids: Optional[List[str]] = None,
        retries: OptionalNullable[utils.RetryConfig] = UNSET,
        server_url: Optional[str] = None,
        timeout_ms: Optional[int] = None,
        http_headers: Optional[Mapping[str, str]] = None,
    ) -> models.DtoCustomerUsageSummaryResponse:
        r"""Get customer usage summary

        Use when showing a customer's usage (e.g. portal or overage alerts). Identify by customer_id or customer_lookup_key; supports filters.

        :param customer_id:
        :param customer_lookup_key:
        :param feature_ids:
        :param feature_lookup_keys:
        :param subscription_ids:
        :param retries: Override the default retry configuration for this method
        :param server_url: Override the default server URL for this method
        :param timeout_ms: Override the default request timeout configuration for this method in milliseconds
        :param http_headers: Additional headers to set or replace on requests.
        """
        base_url = None
        url_variables = None
        if timeout_ms is None:
            timeout_ms = self.sdk_configuration.timeout_ms

        if server_url is not None:
            base_url = server_url
        else:
            base_url = self._get_url(base_url, url_variables)

        request = models.GetCustomerUsageSummaryRequest(
            customer_id=customer_id,
            customer_lookup_key=customer_lookup_key,
            feature_ids=feature_ids,
            feature_lookup_keys=feature_lookup_keys,
            subscription_ids=subscription_ids,
        )

        req = self._build_request(
            method="GET",
            path="/customers/usage",
            base_url=base_url,
            url_variables=url_variables,
            request=request,
            request_body_required=False,
            request_has_path_params=False,
            request_has_query_params=True,
            user_agent_header="user-agent",
            accept_header_value="application/json",
            http_headers=http_headers,
            security=self.sdk_configuration.security,
            allow_empty_value=None,
            timeout_ms=timeout_ms,
        )

        if retries == UNSET:
            if self.sdk_configuration.retry_config is not UNSET:
                retries = self.sdk_configuration.retry_config

        retry_config = None
        if isinstance(retries, utils.RetryConfig):
            retry_config = (retries, ["429", "500", "502", "503", "504"])

        http_res = self.do_request(
            hook_ctx=HookContext(
                config=self.sdk_configuration,
                base_url=base_url or "",
                operation_id="getCustomerUsageSummary",
                oauth2_scopes=None,
                security_source=self.sdk_configuration.security,
            ),
            request=req,
            error_status_codes=["400", "4XX", "500", "5XX"],
            retry_config=retry_config,
        )

        response_data: Any = None
        if utils.match_response(http_res, "200", "application/json"):
            return unmarshal_json_response(
                models.DtoCustomerUsageSummaryResponse, http_res
            )
        if utils.match_response(http_res, "400", "application/json"):
            response_data = unmarshal_json_response(
                models.errors.ErrorsErrorResponseData, http_res
            )
            raise models.errors.ErrorsErrorResponse(response_data, http_res)
        if utils.match_response(http_res, "500", "application/json"):
            response_data = unmarshal_json_response(
                models.errors.ErrorsErrorResponseData, http_res
            )
            raise models.errors.ErrorsErrorResponse(response_data, http_res)
        if utils.match_response(http_res, "4XX", "*"):
            http_res_text = utils.stream_to_text(http_res)
            raise models.errors.FlexpriceDefaultError(
                "API error occurred", http_res, http_res_text
            )
        if utils.match_response(http_res, "5XX", "*"):
            http_res_text = utils.stream_to_text(http_res)
            raise models.errors.FlexpriceDefaultError(
                "API error occurred", http_res, http_res_text
            )

        raise models.errors.FlexpriceDefaultError(
            "Unexpected response received", http_res
        )

    async def get_customer_usage_summary_async(
        self,
        *,
        customer_id: Optional[str] = None,
        customer_lookup_key: Optional[str] = None,
        feature_ids: Optional[List[str]] = None,
        feature_lookup_keys: Optional[List[str]] = None,
        subscription_ids: Optional[List[str]] = None,
        retries: OptionalNullable[utils.RetryConfig] = UNSET,
        server_url: Optional[str] = None,
        timeout_ms: Optional[int] = None,
        http_headers: Optional[Mapping[str, str]] = None,
    ) -> models.DtoCustomerUsageSummaryResponse:
        r"""Get customer usage summary

        Use when showing a customer's usage (e.g. portal or overage alerts). Identify by customer_id or customer_lookup_key; supports filters.

        :param customer_id:
        :param customer_lookup_key:
        :param feature_ids:
        :param feature_lookup_keys:
        :param subscription_ids:
        :param retries: Override the default retry configuration for this method
        :param server_url: Override the default server URL for this method
        :param timeout_ms: Override the default request timeout configuration for this method in milliseconds
        :param http_headers: Additional headers to set or replace on requests.
        """
        base_url = None
        url_variables = None
        if timeout_ms is None:
            timeout_ms = self.sdk_configuration.timeout_ms

        if server_url is not None:
            base_url = server_url
        else:
            base_url = self._get_url(base_url, url_variables)

        request = models.GetCustomerUsageSummaryRequest(
            customer_id=customer_id,
            customer_lookup_key=customer_lookup_key,
            feature_ids=feature_ids,
            feature_lookup_keys=feature_lookup_keys,
            subscription_ids=subscription_ids,
        )

        req = self._build_request_async(
            method="GET",
            path="/customers/usage",
            base_url=base_url,
            url_variables=url_variables,
            request=request,
            request_body_required=False,
            request_has_path_params=False,
            request_has_query_params=True,
            user_agent_header="user-agent",
            accept_header_value="application/json",
            http_headers=http_headers,
            security=self.sdk_configuration.security,
            allow_empty_value=None,
            timeout_ms=timeout_ms,
        )

        if retries == UNSET:
            if self.sdk_configuration.retry_config is not UNSET:
                retries = self.sdk_configuration.retry_config

        retry_config = None
        if isinstance(retries, utils.RetryConfig):
            retry_config = (retries, ["429", "500", "502", "503", "504"])

        http_res = await self.do_request_async(
            hook_ctx=HookContext(
                config=self.sdk_configuration,
                base_url=base_url or "",
                operation_id="getCustomerUsageSummary",
                oauth2_scopes=None,
                security_source=self.sdk_configuration.security,
            ),
            request=req,
            error_status_codes=["400", "4XX", "500", "5XX"],
            retry_config=retry_config,
        )

        response_data: Any = None
        if utils.match_response(http_res, "200", "application/json"):
            return unmarshal_json_response(
                models.DtoCustomerUsageSummaryResponse, http_res
            )
        if utils.match_response(http_res, "400", "application/json"):
            response_data = unmarshal_json_response(
                models.errors.ErrorsErrorResponseData, http_res
            )
            raise models.errors.ErrorsErrorResponse(response_data, http_res)
        if utils.match_response(http_res, "500", "application/json"):
            response_data = unmarshal_json_response(
                models.errors.ErrorsErrorResponseData, http_res
            )
            raise models.errors.ErrorsErrorResponse(response_data, http_res)
        if utils.match_response(http_res, "4XX", "*"):
            http_res_text = await utils.stream_to_text_async(http_res)
            raise models.errors.FlexpriceDefaultError(
                "API error occurred", http_res, http_res_text
            )
        if utils.match_response(http_res, "5XX", "*"):
            http_res_text = await utils.stream_to_text_async(http_res)
            raise models.errors.FlexpriceDefaultError(
                "API error occurred", http_res, http_res_text
            )

        raise models.errors.FlexpriceDefaultError(
            "Unexpected response received", http_res
        )

    def get_customer(
        self,
        *,
        id: str,
        retries: OptionalNullable[utils.RetryConfig] = UNSET,
        server_url: Optional[str] = None,
        timeout_ms: Optional[int] = None,
        http_headers: Optional[Mapping[str, str]] = None,
    ) -> models.DtoCustomerResponse:
        r"""Get customer

        Use when you need to load a single customer (e.g. for a billing portal or to attach a subscription).

        :param id: Customer ID
        :param retries: Override the default retry configuration for this method
        :param server_url: Override the default server URL for this method
        :param timeout_ms: Override the default request timeout configuration for this method in milliseconds
        :param http_headers: Additional headers to set or replace on requests.
        """
        base_url = None
        url_variables = None
        if timeout_ms is None:
            timeout_ms = self.sdk_configuration.timeout_ms

        if server_url is not None:
            base_url = server_url
        else:
            base_url = self._get_url(base_url, url_variables)

        request = models.GetCustomerRequest(
            id=id,
        )

        req = self._build_request(
            method="GET",
            path="/customers/{id}",
            base_url=base_url,
            url_variables=url_variables,
            request=request,
            request_body_required=False,
            request_has_path_params=True,
            request_has_query_params=True,
            user_agent_header="user-agent",
            accept_header_value="application/json",
            http_headers=http_headers,
            security=self.sdk_configuration.security,
            allow_empty_value=None,
            timeout_ms=timeout_ms,
        )

        if retries == UNSET:
            if self.sdk_configuration.retry_config is not UNSET:
                retries = self.sdk_configuration.retry_config

        retry_config = None
        if isinstance(retries, utils.RetryConfig):
            retry_config = (retries, ["429", "500", "502", "503", "504"])

        http_res = self.do_request(
            hook_ctx=HookContext(
                config=self.sdk_configuration,
                base_url=base_url or "",
                operation_id="getCustomer",
                oauth2_scopes=None,
                security_source=self.sdk_configuration.security,
            ),
            request=req,
            error_status_codes=["400", "4XX", "500", "5XX"],
            retry_config=retry_config,
        )

        response_data: Any = None
        if utils.match_response(http_res, "200", "application/json"):
            return unmarshal_json_response(models.DtoCustomerResponse, http_res)
        if utils.match_response(http_res, "400", "application/json"):
            response_data = unmarshal_json_response(
                models.errors.ErrorsErrorResponseData, http_res
            )
            raise models.errors.ErrorsErrorResponse(response_data, http_res)
        if utils.match_response(http_res, "500", "application/json"):
            response_data = unmarshal_json_response(
                models.errors.ErrorsErrorResponseData, http_res
            )
            raise models.errors.ErrorsErrorResponse(response_data, http_res)
        if utils.match_response(http_res, "4XX", "*"):
            http_res_text = utils.stream_to_text(http_res)
            raise models.errors.FlexpriceDefaultError(
                "API error occurred", http_res, http_res_text
            )
        if utils.match_response(http_res, "5XX", "*"):
            http_res_text = utils.stream_to_text(http_res)
            raise models.errors.FlexpriceDefaultError(
                "API error occurred", http_res, http_res_text
            )

        raise models.errors.FlexpriceDefaultError(
            "Unexpected response received", http_res
        )

    async def get_customer_async(
        self,
        *,
        id: str,
        retries: OptionalNullable[utils.RetryConfig] = UNSET,
        server_url: Optional[str] = None,
        timeout_ms: Optional[int] = None,
        http_headers: Optional[Mapping[str, str]] = None,
    ) -> models.DtoCustomerResponse:
        r"""Get customer

        Use when you need to load a single customer (e.g. for a billing portal or to attach a subscription).

        :param id: Customer ID
        :param retries: Override the default retry configuration for this method
        :param server_url: Override the default server URL for this method
        :param timeout_ms: Override the default request timeout configuration for this method in milliseconds
        :param http_headers: Additional headers to set or replace on requests.
        """
        base_url = None
        url_variables = None
        if timeout_ms is None:
            timeout_ms = self.sdk_configuration.timeout_ms

        if server_url is not None:
            base_url = server_url
        else:
            base_url = self._get_url(base_url, url_variables)

        request = models.GetCustomerRequest(
            id=id,
        )

        req = self._build_request_async(
            method="GET",
            path="/customers/{id}",
            base_url=base_url,
            url_variables=url_variables,
            request=request,
            request_body_required=False,
            request_has_path_params=True,
            request_has_query_params=True,
            user_agent_header="user-agent",
            accept_header_value="application/json",
            http_headers=http_headers,
            security=self.sdk_configuration.security,
            allow_empty_value=None,
            timeout_ms=timeout_ms,
        )

        if retries == UNSET:
            if self.sdk_configuration.retry_config is not UNSET:
                retries = self.sdk_configuration.retry_config

        retry_config = None
        if isinstance(retries, utils.RetryConfig):
            retry_config = (retries, ["429", "500", "502", "503", "504"])

        http_res = await self.do_request_async(
            hook_ctx=HookContext(
                config=self.sdk_configuration,
                base_url=base_url or "",
                operation_id="getCustomer",
                oauth2_scopes=None,
                security_source=self.sdk_configuration.security,
            ),
            request=req,
            error_status_codes=["400", "4XX", "500", "5XX"],
            retry_config=retry_config,
        )

        response_data: Any = None
        if utils.match_response(http_res, "200", "application/json"):
            return unmarshal_json_response(models.DtoCustomerResponse, http_res)
        if utils.match_response(http_res, "400", "application/json"):
            response_data = unmarshal_json_response(
                models.errors.ErrorsErrorResponseData, http_res
            )
            raise models.errors.ErrorsErrorResponse(response_data, http_res)
        if utils.match_response(http_res, "500", "application/json"):
            response_data = unmarshal_json_response(
                models.errors.ErrorsErrorResponseData, http_res
            )
            raise models.errors.ErrorsErrorResponse(response_data, http_res)
        if utils.match_response(http_res, "4XX", "*"):
            http_res_text = await utils.stream_to_text_async(http_res)
            raise models.errors.FlexpriceDefaultError(
                "API error occurred", http_res, http_res_text
            )
        if utils.match_response(http_res, "5XX", "*"):
            http_res_text = await utils.stream_to_text_async(http_res)
            raise models.errors.FlexpriceDefaultError(
                "API error occurred", http_res, http_res_text
            )

        raise models.errors.FlexpriceDefaultError(
            "Unexpected response received", http_res
        )

    def delete_customer(
        self,
        *,
        id: str,
        retries: OptionalNullable[utils.RetryConfig] = UNSET,
        server_url: Optional[str] = None,
        timeout_ms: Optional[int] = None,
        http_headers: Optional[Mapping[str, str]] = None,
    ):
        r"""Delete customer

        Use when removing a customer (e.g. GDPR or churn). Returns 204 No Content on success.

        :param id: Customer ID
        :param retries: Override the default retry configuration for this method
        :param server_url: Override the default server URL for this method
        :param timeout_ms: Override the default request timeout configuration for this method in milliseconds
        :param http_headers: Additional headers to set or replace on requests.
        """
        base_url = None
        url_variables = None
        if timeout_ms is None:
            timeout_ms = self.sdk_configuration.timeout_ms

        if server_url is not None:
            base_url = server_url
        else:
            base_url = self._get_url(base_url, url_variables)

        request = models.DeleteCustomerRequest(
            id=id,
        )

        req = self._build_request(
            method="DELETE",
            path="/customers/{id}",
            base_url=base_url,
            url_variables=url_variables,
            request=request,
            request_body_required=False,
            request_has_path_params=True,
            request_has_query_params=True,
            user_agent_header="user-agent",
            accept_header_value="application/json",
            http_headers=http_headers,
            security=self.sdk_configuration.security,
            allow_empty_value=None,
            timeout_ms=timeout_ms,
        )

        if retries == UNSET:
            if self.sdk_configuration.retry_config is not UNSET:
                retries = self.sdk_configuration.retry_config

        retry_config = None
        if isinstance(retries, utils.RetryConfig):
            retry_config = (retries, ["429", "500", "502", "503", "504"])

        http_res = self.do_request(
            hook_ctx=HookContext(
                config=self.sdk_configuration,
                base_url=base_url or "",
                operation_id="deleteCustomer",
                oauth2_scopes=None,
                security_source=self.sdk_configuration.security,
            ),
            request=req,
            error_status_codes=["400", "4XX", "500", "5XX"],
            retry_config=retry_config,
        )

        response_data: Any = None
        if utils.match_response(http_res, "204", "*"):
            return
        if utils.match_response(http_res, "400", "application/json"):
            response_data = unmarshal_json_response(
                models.errors.ErrorsErrorResponseData, http_res
            )
            raise models.errors.ErrorsErrorResponse(response_data, http_res)
        if utils.match_response(http_res, "500", "application/json"):
            response_data = unmarshal_json_response(
                models.errors.ErrorsErrorResponseData, http_res
            )
            raise models.errors.ErrorsErrorResponse(response_data, http_res)
        if utils.match_response(http_res, "4XX", "*"):
            http_res_text = utils.stream_to_text(http_res)
            raise models.errors.FlexpriceDefaultError(
                "API error occurred", http_res, http_res_text
            )
        if utils.match_response(http_res, "5XX", "*"):
            http_res_text = utils.stream_to_text(http_res)
            raise models.errors.FlexpriceDefaultError(
                "API error occurred", http_res, http_res_text
            )

        raise models.errors.FlexpriceDefaultError(
            "Unexpected response received", http_res
        )

    async def delete_customer_async(
        self,
        *,
        id: str,
        retries: OptionalNullable[utils.RetryConfig] = UNSET,
        server_url: Optional[str] = None,
        timeout_ms: Optional[int] = None,
        http_headers: Optional[Mapping[str, str]] = None,
    ):
        r"""Delete customer

        Use when removing a customer (e.g. GDPR or churn). Returns 204 No Content on success.

        :param id: Customer ID
        :param retries: Override the default retry configuration for this method
        :param server_url: Override the default server URL for this method
        :param timeout_ms: Override the default request timeout configuration for this method in milliseconds
        :param http_headers: Additional headers to set or replace on requests.
        """
        base_url = None
        url_variables = None
        if timeout_ms is None:
            timeout_ms = self.sdk_configuration.timeout_ms

        if server_url is not None:
            base_url = server_url
        else:
            base_url = self._get_url(base_url, url_variables)

        request = models.DeleteCustomerRequest(
            id=id,
        )

        req = self._build_request_async(
            method="DELETE",
            path="/customers/{id}",
            base_url=base_url,
            url_variables=url_variables,
            request=request,
            request_body_required=False,
            request_has_path_params=True,
            request_has_query_params=True,
            user_agent_header="user-agent",
            accept_header_value="application/json",
            http_headers=http_headers,
            security=self.sdk_configuration.security,
            allow_empty_value=None,
            timeout_ms=timeout_ms,
        )

        if retries == UNSET:
            if self.sdk_configuration.retry_config is not UNSET:
                retries = self.sdk_configuration.retry_config

        retry_config = None
        if isinstance(retries, utils.RetryConfig):
            retry_config = (retries, ["429", "500", "502", "503", "504"])

        http_res = await self.do_request_async(
            hook_ctx=HookContext(
                config=self.sdk_configuration,
                base_url=base_url or "",
                operation_id="deleteCustomer",
                oauth2_scopes=None,
                security_source=self.sdk_configuration.security,
            ),
            request=req,
            error_status_codes=["400", "4XX", "500", "5XX"],
            retry_config=retry_config,
        )

        response_data: Any = None
        if utils.match_response(http_res, "204", "*"):
            return
        if utils.match_response(http_res, "400", "application/json"):
            response_data = unmarshal_json_response(
                models.errors.ErrorsErrorResponseData, http_res
            )
            raise models.errors.ErrorsErrorResponse(response_data, http_res)
        if utils.match_response(http_res, "500", "application/json"):
            response_data = unmarshal_json_response(
                models.errors.ErrorsErrorResponseData, http_res
            )
            raise models.errors.ErrorsErrorResponse(response_data, http_res)
        if utils.match_response(http_res, "4XX", "*"):
            http_res_text = await utils.stream_to_text_async(http_res)
            raise models.errors.FlexpriceDefaultError(
                "API error occurred", http_res, http_res_text
            )
        if utils.match_response(http_res, "5XX", "*"):
            http_res_text = await utils.stream_to_text_async(http_res)
            raise models.errors.FlexpriceDefaultError(
                "API error occurred", http_res, http_res_text
            )

        raise models.errors.FlexpriceDefaultError(
            "Unexpected response received", http_res
        )

    def get_customer_entitlements(
        self,
        *,
        id: str,
        retries: OptionalNullable[utils.RetryConfig] = UNSET,
        server_url: Optional[str] = None,
        timeout_ms: Optional[int] = None,
        http_headers: Optional[Mapping[str, str]] = None,
    ) -> models.DtoCustomerEntitlementsResponse:
        r"""Get customer entitlements

        Use when checking what a customer can access (e.g. feature gating or usage limits). Supports optional filters (feature_ids, subscription_ids).

        :param id: Customer ID
        :param retries: Override the default retry configuration for this method
        :param server_url: Override the default server URL for this method
        :param timeout_ms: Override the default request timeout configuration for this method in milliseconds
        :param http_headers: Additional headers to set or replace on requests.
        """
        base_url = None
        url_variables = None
        if timeout_ms is None:
            timeout_ms = self.sdk_configuration.timeout_ms

        if server_url is not None:
            base_url = server_url
        else:
            base_url = self._get_url(base_url, url_variables)

        request = models.GetCustomerEntitlementsRequest(
            id=id,
        )

        req = self._build_request(
            method="GET",
            path="/customers/{id}/entitlements",
            base_url=base_url,
            url_variables=url_variables,
            request=request,
            request_body_required=False,
            request_has_path_params=True,
            request_has_query_params=True,
            user_agent_header="user-agent",
            accept_header_value="application/json",
            http_headers=http_headers,
            security=self.sdk_configuration.security,
            allow_empty_value=None,
            timeout_ms=timeout_ms,
        )

        if retries == UNSET:
            if self.sdk_configuration.retry_config is not UNSET:
                retries = self.sdk_configuration.retry_config

        retry_config = None
        if isinstance(retries, utils.RetryConfig):
            retry_config = (retries, ["429", "500", "502", "503", "504"])

        http_res = self.do_request(
            hook_ctx=HookContext(
                config=self.sdk_configuration,
                base_url=base_url or "",
                operation_id="getCustomerEntitlements",
                oauth2_scopes=None,
                security_source=self.sdk_configuration.security,
            ),
            request=req,
            error_status_codes=["400", "4XX", "500", "5XX"],
            retry_config=retry_config,
        )

        response_data: Any = None
        if utils.match_response(http_res, "200", "application/json"):
            return unmarshal_json_response(
                models.DtoCustomerEntitlementsResponse, http_res
            )
        if utils.match_response(http_res, "400", "application/json"):
            response_data = unmarshal_json_response(
                models.errors.ErrorsErrorResponseData, http_res
            )
            raise models.errors.ErrorsErrorResponse(response_data, http_res)
        if utils.match_response(http_res, "500", "application/json"):
            response_data = unmarshal_json_response(
                models.errors.ErrorsErrorResponseData, http_res
            )
            raise models.errors.ErrorsErrorResponse(response_data, http_res)
        if utils.match_response(http_res, "4XX", "*"):
            http_res_text = utils.stream_to_text(http_res)
            raise models.errors.FlexpriceDefaultError(
                "API error occurred", http_res, http_res_text
            )
        if utils.match_response(http_res, "5XX", "*"):
            http_res_text = utils.stream_to_text(http_res)
            raise models.errors.FlexpriceDefaultError(
                "API error occurred", http_res, http_res_text
            )

        raise models.errors.FlexpriceDefaultError(
            "Unexpected response received", http_res
        )

    async def get_customer_entitlements_async(
        self,
        *,
        id: str,
        retries: OptionalNullable[utils.RetryConfig] = UNSET,
        server_url: Optional[str] = None,
        timeout_ms: Optional[int] = None,
        http_headers: Optional[Mapping[str, str]] = None,
    ) -> models.DtoCustomerEntitlementsResponse:
        r"""Get customer entitlements

        Use when checking what a customer can access (e.g. feature gating or usage limits). Supports optional filters (feature_ids, subscription_ids).

        :param id: Customer ID
        :param retries: Override the default retry configuration for this method
        :param server_url: Override the default server URL for this method
        :param timeout_ms: Override the default request timeout configuration for this method in milliseconds
        :param http_headers: Additional headers to set or replace on requests.
        """
        base_url = None
        url_variables = None
        if timeout_ms is None:
            timeout_ms = self.sdk_configuration.timeout_ms

        if server_url is not None:
            base_url = server_url
        else:
            base_url = self._get_url(base_url, url_variables)

        request = models.GetCustomerEntitlementsRequest(
            id=id,
        )

        req = self._build_request_async(
            method="GET",
            path="/customers/{id}/entitlements",
            base_url=base_url,
            url_variables=url_variables,
            request=request,
            request_body_required=False,
            request_has_path_params=True,
            request_has_query_params=True,
            user_agent_header="user-agent",
            accept_header_value="application/json",
            http_headers=http_headers,
            security=self.sdk_configuration.security,
            allow_empty_value=None,
            timeout_ms=timeout_ms,
        )

        if retries == UNSET:
            if self.sdk_configuration.retry_config is not UNSET:
                retries = self.sdk_configuration.retry_config

        retry_config = None
        if isinstance(retries, utils.RetryConfig):
            retry_config = (retries, ["429", "500", "502", "503", "504"])

        http_res = await self.do_request_async(
            hook_ctx=HookContext(
                config=self.sdk_configuration,
                base_url=base_url or "",
                operation_id="getCustomerEntitlements",
                oauth2_scopes=None,
                security_source=self.sdk_configuration.security,
            ),
            request=req,
            error_status_codes=["400", "4XX", "500", "5XX"],
            retry_config=retry_config,
        )

        response_data: Any = None
        if utils.match_response(http_res, "200", "application/json"):
            return unmarshal_json_response(
                models.DtoCustomerEntitlementsResponse, http_res
            )
        if utils.match_response(http_res, "400", "application/json"):
            response_data = unmarshal_json_response(
                models.errors.ErrorsErrorResponseData, http_res
            )
            raise models.errors.ErrorsErrorResponse(response_data, http_res)
        if utils.match_response(http_res, "500", "application/json"):
            response_data = unmarshal_json_response(
                models.errors.ErrorsErrorResponseData, http_res
            )
            raise models.errors.ErrorsErrorResponse(response_data, http_res)
        if utils.match_response(http_res, "4XX", "*"):
            http_res_text = await utils.stream_to_text_async(http_res)
            raise models.errors.FlexpriceDefaultError(
                "API error occurred", http_res, http_res_text
            )
        if utils.match_response(http_res, "5XX", "*"):
            http_res_text = await utils.stream_to_text_async(http_res)
            raise models.errors.FlexpriceDefaultError(
                "API error occurred", http_res, http_res_text
            )

        raise models.errors.FlexpriceDefaultError(
            "Unexpected response received", http_res
        )

    def get_customer_upcoming_grants(
        self,
        *,
        id: str,
        retries: OptionalNullable[utils.RetryConfig] = UNSET,
        server_url: Optional[str] = None,
        timeout_ms: Optional[int] = None,
        http_headers: Optional[Mapping[str, str]] = None,
    ) -> models.DtoListCreditGrantApplicationsResponse:
        r"""Get upcoming credit grant applications

        Use when showing upcoming or pending credits for a customer (e.g. in a portal or for forecasting).

        :param id: Customer ID
        :param retries: Override the default retry configuration for this method
        :param server_url: Override the default server URL for this method
        :param timeout_ms: Override the default request timeout configuration for this method in milliseconds
        :param http_headers: Additional headers to set or replace on requests.
        """
        base_url = None
        url_variables = None
        if timeout_ms is None:
            timeout_ms = self.sdk_configuration.timeout_ms

        if server_url is not None:
            base_url = server_url
        else:
            base_url = self._get_url(base_url, url_variables)

        request = models.GetCustomerUpcomingGrantsRequest(
            id=id,
        )

        req = self._build_request(
            method="GET",
            path="/customers/{id}/grants/upcoming",
            base_url=base_url,
            url_variables=url_variables,
            request=request,
            request_body_required=False,
            request_has_path_params=True,
            request_has_query_params=True,
            user_agent_header="user-agent",
            accept_header_value="application/json",
            http_headers=http_headers,
            security=self.sdk_configuration.security,
            allow_empty_value=None,
            timeout_ms=timeout_ms,
        )

        if retries == UNSET:
            if self.sdk_configuration.retry_config is not UNSET:
                retries = self.sdk_configuration.retry_config

        retry_config = None
        if isinstance(retries, utils.RetryConfig):
            retry_config = (retries, ["429", "500", "502", "503", "504"])

        http_res = self.do_request(
            hook_ctx=HookContext(
                config=self.sdk_configuration,
                base_url=base_url or "",
                operation_id="getCustomerUpcomingGrants",
                oauth2_scopes=None,
                security_source=self.sdk_configuration.security,
            ),
            request=req,
            error_status_codes=["400", "404", "4XX", "500", "5XX"],
            retry_config=retry_config,
        )

        response_data: Any = None
        if utils.match_response(http_res, "200", "application/json"):
            return unmarshal_json_response(
                models.DtoListCreditGrantApplicationsResponse, http_res
            )
        if utils.match_response(http_res, ["400", "404"], "application/json"):
            response_data = unmarshal_json_response(
                models.errors.ErrorsErrorResponseData, http_res
            )
            raise models.errors.ErrorsErrorResponse(response_data, http_res)
        if utils.match_response(http_res, "500", "application/json"):
            response_data = unmarshal_json_response(
                models.errors.ErrorsErrorResponseData, http_res
            )
            raise models.errors.ErrorsErrorResponse(response_data, http_res)
        if utils.match_response(http_res, "4XX", "*"):
            http_res_text = utils.stream_to_text(http_res)
            raise models.errors.FlexpriceDefaultError(
                "API error occurred", http_res, http_res_text
            )
        if utils.match_response(http_res, "5XX", "*"):
            http_res_text = utils.stream_to_text(http_res)
            raise models.errors.FlexpriceDefaultError(
                "API error occurred", http_res, http_res_text
            )

        raise models.errors.FlexpriceDefaultError(
            "Unexpected response received", http_res
        )

    async def get_customer_upcoming_grants_async(
        self,
        *,
        id: str,
        retries: OptionalNullable[utils.RetryConfig] = UNSET,
        server_url: Optional[str] = None,
        timeout_ms: Optional[int] = None,
        http_headers: Optional[Mapping[str, str]] = None,
    ) -> models.DtoListCreditGrantApplicationsResponse:
        r"""Get upcoming credit grant applications

        Use when showing upcoming or pending credits for a customer (e.g. in a portal or for forecasting).

        :param id: Customer ID
        :param retries: Override the default retry configuration for this method
        :param server_url: Override the default server URL for this method
        :param timeout_ms: Override the default request timeout configuration for this method in milliseconds
        :param http_headers: Additional headers to set or replace on requests.
        """
        base_url = None
        url_variables = None
        if timeout_ms is None:
            timeout_ms = self.sdk_configuration.timeout_ms

        if server_url is not None:
            base_url = server_url
        else:
            base_url = self._get_url(base_url, url_variables)

        request = models.GetCustomerUpcomingGrantsRequest(
            id=id,
        )

        req = self._build_request_async(
            method="GET",
            path="/customers/{id}/grants/upcoming",
            base_url=base_url,
            url_variables=url_variables,
            request=request,
            request_body_required=False,
            request_has_path_params=True,
            request_has_query_params=True,
            user_agent_header="user-agent",
            accept_header_value="application/json",
            http_headers=http_headers,
            security=self.sdk_configuration.security,
            allow_empty_value=None,
            timeout_ms=timeout_ms,
        )

        if retries == UNSET:
            if self.sdk_configuration.retry_config is not UNSET:
                retries = self.sdk_configuration.retry_config

        retry_config = None
        if isinstance(retries, utils.RetryConfig):
            retry_config = (retries, ["429", "500", "502", "503", "504"])

        http_res = await self.do_request_async(
            hook_ctx=HookContext(
                config=self.sdk_configuration,
                base_url=base_url or "",
                operation_id="getCustomerUpcomingGrants",
                oauth2_scopes=None,
                security_source=self.sdk_configuration.security,
            ),
            request=req,
            error_status_codes=["400", "404", "4XX", "500", "5XX"],
            retry_config=retry_config,
        )

        response_data: Any = None
        if utils.match_response(http_res, "200", "application/json"):
            return unmarshal_json_response(
                models.DtoListCreditGrantApplicationsResponse, http_res
            )
        if utils.match_response(http_res, ["400", "404"], "application/json"):
            response_data = unmarshal_json_response(
                models.errors.ErrorsErrorResponseData, http_res
            )
            raise models.errors.ErrorsErrorResponse(response_data, http_res)
        if utils.match_response(http_res, "500", "application/json"):
            response_data = unmarshal_json_response(
                models.errors.ErrorsErrorResponseData, http_res
            )
            raise models.errors.ErrorsErrorResponse(response_data, http_res)
        if utils.match_response(http_res, "4XX", "*"):
            http_res_text = await utils.stream_to_text_async(http_res)
            raise models.errors.FlexpriceDefaultError(
                "API error occurred", http_res, http_res_text
            )
        if utils.match_response(http_res, "5XX", "*"):
            http_res_text = await utils.stream_to_text_async(http_res)
            raise models.errors.FlexpriceDefaultError(
                "API error occurred", http_res, http_res_text
            )

        raise models.errors.FlexpriceDefaultError(
            "Unexpected response received", http_res
        )



"""Code generated by Speakeasy (https://speakeasy.com). DO NOT EDIT."""

from __future__ import annotations
from .price_price import PricePrice, PricePriceTypedDict
from flexprice.types import BaseModel, UNSET_SENTINEL
from pydantic import model_serializer
from typing import Dict, List, Optional
from typing_extensions import NotRequired, TypedDict


class DtoSubscriptionUsageByMetersResponseTypedDict(TypedDict):
    amount: NotRequired[float]
    currency: NotRequired[str]
    display_amount: NotRequired[str]
    filter_values: NotRequired[Dict[str, List[str]]]
    is_overage: NotRequired[bool]
    r"""Whether this charge is at overage rate"""
    meter_display_name: NotRequired[str]
    meter_id: NotRequired[str]
    overage_factor: NotRequired[float]
    r"""Factor applied to this charge if in overage"""
    price: NotRequired[PricePriceTypedDict]
    quantity: NotRequired[float]
    subscription_line_item_id: NotRequired[str]
    r"""For feature_usage: direct match by sub_line_item_id"""


class DtoSubscriptionUsageByMetersResponse(BaseModel):
    amount: Optional[float] = None

    currency: Optional[str] = None

    display_amount: Optional[str] = None

    filter_values: Optional[Dict[str, List[str]]] = None

    is_overage: Optional[bool] = None
    r"""Whether this charge is at overage rate"""

    meter_display_name: Optional[str] = None

    meter_id: Optional[str] = None

    overage_factor: Optional[float] = None
    r"""Factor applied to this charge if in overage"""

    price: Optional[PricePrice] = None

    quantity: Optional[float] = None

    subscription_line_item_id: Optional[str] = None
    r"""For feature_usage: direct match by sub_line_item_id"""

    @model_serializer(mode="wrap")
    def serialize_model(self, handler):
        optional_fields = set(
            [
                "amount",
                "currency",
                "display_amount",
                "filter_values",
                "is_overage",
                "meter_display_name",
                "meter_id",
                "overage_factor",
                "price",
                "quantity",
                "subscription_line_item_id",
            ]
        )
        serialized = handler(self)
        m = {}

        for n, f in type(self).model_fields.items():
            k = f.alias or n
            val = serialized.get(k, serialized.get(n))

            if val != UNSET_SENTINEL:
                if val is not None or k not in optional_fields:
                    m[k] = val

        return m
